# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
from pathlib import Path
from typing import Dict, Optional

import numpy as np
import tiktoken
import torch

PATTERN_TIKTOKEN = "[^\\r\\n\\p{L}\\p{N}]?[\\p{Lu}\\p{Lt}\\p{Lm}\\p{Lo}\\p{M}]*[\\p{Ll}\\p{Lm}\\p{Lo}\\p{M}]+|[^\\r\\n\\p{L}\\p{N}]?[\\p{Lu}\\p{Lt}\\p{Lm}\\p{Lo}\\p{M}]+[\\p{Ll}\\p{Lm}\\p{Lo}\\p{M}]*|\\p{N}| ?[^\\s\\p{L}\\p{N}]+[\\r\\n/]*|\\s*[\\r\\n]+|\\s+(?!\\S)|\\s+"
DEFAULT_TIKTOKEN_MAX_VOCAB = 2**17  # 131072
SPECIAL_TOKENS = ["<unk>", "<s>", "</s>"]
SPECIAL_TOKEN_TEMPLATE = "<SPECIAL_{id}>"


def reload_mergeable_ranks(
    path: str,
    max_vocab: Optional[int] = None,
) -> Dict[bytes, int]:
    """
    Reload the tokenizer JSON file and convert it to Tiktoken format.
    """
    assert path.endswith(".json")

    # reload vocab
    with open(path, "r", encoding='utf-8') as f:
        vocab = json.load(f)
    assert isinstance(vocab, list)
    print(f"Vocab size: {len(vocab)}")
    if max_vocab is not None:
        vocab = vocab[:max_vocab]
        print(f"Cutting vocab to first {len(vocab)} tokens.")

    # build ranks
    ranks: Dict[bytes, int] = {}
    for i, x in enumerate(vocab):
        assert x.keys() == {"rank", "token_bytes", "token_str"}
        assert x["rank"] == i
        merge = base64.b64decode(x["token_bytes"])
        assert i >= 256 or merge == bytes([i])
        ranks[merge] = x["rank"]

    # sanity check
    assert len(ranks) == len(vocab)
    assert set(ranks.values()) == set(range(len(ranks)))

    return ranks


class TiktokenTokenizer:
    def __init__(self, vocab_file: str):

        self.num_special_tokens = 1000
        vocab_size = DEFAULT_TIKTOKEN_MAX_VOCAB
        pattern = PATTERN_TIKTOKEN
        special_tokens = SPECIAL_TOKENS.copy()
        inner_vocab_size = vocab_size - self.num_special_tokens

        token2id = reload_mergeable_ranks(vocab_file, max_vocab=inner_vocab_size)
        self.tokenizer = tiktoken.Encoding(
            name=Path(vocab_file).parent.name,
            pat_str=pattern,
            mergeable_ranks=token2id,
            special_tokens={},  # special tokens are handled manually
        )

        # BOS / EOS / Pad token IDs
        self._bos_id = special_tokens.index("<s>")
        self._eos_id = special_tokens.index("</s>")

    def encode(self, text):
        tokens = self.tokenizer.encode(text)
        tokens = [t + self.num_special_tokens for t in tokens]
        return tokens

    def decode(self, tokens):
        # Filter out special tokens and adjust the remaining tokens
        adjusted_tokens = [
            t - self.num_special_tokens
            for t in tokens
            if t not in {self._bos_id, self._eos_id} and t >= self.num_special_tokens
        ]

        # Decode only if there are tokens left after filtering
        if adjusted_tokens:
            return self.tokenizer.decode(adjusted_tokens)
        else:
            return ""  # Return an empty string if all tokens were filtered out

    def batch_decode(self, ids):
        if isinstance(ids, np.ndarray) or torch.is_tensor(ids):
            ids = ids.tolist()

        if isinstance(ids[0], list):
            ids = ids[0]

        return self.decode(ids)

    @property
    def pad_id(self):
        return self._eos_id

    @property
    def bos_token_id(self):
        return self._bos_id

    @property
    def eos_token_id(self):
        return self._eos_id



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import lightning.pytorch as pl
import nemo_run as run
import torch
from lightning.pytorch.callbacks.callback import Callback

from nemo import lightning as nl
from nemo.collections.llm.api import finetune, pretrain
from nemo.collections.llm.gpt.data.mock import MockDataModule
from nemo.collections.llm.gpt.model.starcoder import StarcoderConfig15B, StarcoderModel
from nemo.collections.llm.peft import PEFT_STR2CLS
from nemo.collections.llm.recipes.finetune_default import default_finetune_recipe
from nemo.collections.llm.recipes.log.default import default_log, default_resume, tensorboard_logger
from nemo.collections.llm.recipes.optim.adam import distributed_fused_adam_with_cosine_annealing
from nemo.collections.llm.recipes.precision.mixed_precision import bf16_mixed, fp16_mixed
from nemo.utils.exp_manager import TimingCallback

NAME = "starcoder_15b"


@run.cli.factory(name=NAME)
def model() -> run.Config[pl.LightningModule]:
    """
    Factory function to create a Starcoder 15B model configuration.

    Returns:
        run.Config[pl.LightningModule]: Configuration for the Starcoder 15B model.

    Examples:
        CLI usage:
            $ nemo llm pretrain model=starcoder_15b ...

        Python API usage:
            >>> model_config = model()
            >>> print(model_config)
    """

    return run.Config(StarcoderModel, config=run.Config(StarcoderConfig15B))


def starcoder_trainer(
    tensor_parallelism: int = 4,
    pipeline_parallelism: int = 2,
    pipeline_parallelism_type: Optional[torch.dtype] = None,
    virtual_pipeline_parallelism: Optional[int] = None,
    context_parallelism: int = 1,
    sequence_parallelism: bool = False,
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    max_steps: int = 1168251,
    precision: str = "bf16-mixed",
    accumulate_grad_batches: int = 1,
    limit_test_batches: int = 32,
    limit_val_batches: int = 32,
    log_every_n_steps: int = 10,
    val_check_interval: int = 2000,
    callbacks: Optional[list[run.Config[Callback]]] = None,
) -> run.Config[nl.Trainer]:
    """
    Configure the NeMo Lightning Trainer for Starcoder 15B models.

    This function sets up the distributed training strategy and other training parameters.

    Args:
        tensor_parallelism (int): Degree of tensor model parallelism.
        pipeline_parallelism (int): Degree of pipeline model parallelism.
        pipeline_parallelism_type (Optional[torch.dtype]): Data type for pipeline parallelism.
        virtual_pipeline_parallelism (Optional[int]): Size of virtual pipeline parallelism.
        context_parallelism (int): Degree of context parallelism.
        sequence_parallelism (bool): Whether to use sequence parallelism.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        max_steps (int): Maximum number of training steps.
        precision (str): Precision configuration, one of fp32, 16-mixed or bf16-mixed.
        accumulate_grad_batches (int): Number of steps per gradient accumulation.
        limit_test_batches (int): Limit the number of test batches.
        limit_val_batches (int): Limit the number of validation batches.
        log_every_n_steps (int): Log every n steps.
        val_check_interval (int): Run validation every N steps.
        callbacks (Optional[list[run.Config[Callback]]]): List of callback configurations.

    Returns:
        run.Config[nl.Trainer]: Configuration for the NeMo Lightning Trainer.
    """
    strategy = run.Config(
        nl.MegatronStrategy,
        tensor_model_parallel_size=tensor_parallelism,
        pipeline_model_parallel_size=pipeline_parallelism,
        pipeline_dtype=pipeline_parallelism_type,
        virtual_pipeline_model_parallel_size=virtual_pipeline_parallelism,
        context_parallel_size=context_parallelism,
        sequence_parallel=sequence_parallelism,
        gradient_as_bucket_view=True,
        ckpt_include_optimizer=True,
        ckpt_async_save=True,
        ckpt_parallel_load=True,
    )

    precision_plugin = None
    if precision == "16-mixed":
        precision_plugin = fp16_mixed()
    elif precision == "bf16-mixed":
        precision_plugin = bf16_mixed()

    trainer = run.Config(
        nl.Trainer,
        accelerator="gpu",
        callbacks=callbacks,
        devices=num_gpus_per_node,
        accumulate_grad_batches=accumulate_grad_batches,
        limit_test_batches=limit_test_batches,
        limit_val_batches=limit_val_batches,
        log_every_n_steps=log_every_n_steps,
        max_steps=max_steps,
        num_nodes=num_nodes,
        plugins=precision_plugin,
        strategy=strategy,
        use_distributed_sampler=False,
        val_check_interval=val_check_interval,
    )

    return trainer


@run.cli.factory(target=pretrain, name=NAME)
def pretrain_recipe(
    # General
    dir: Optional[str] = None,
    name: str = "default",
    # Trainer
    tensor_parallelism: int = 1,
    pipeline_parallelism: int = 8,
    pipeline_parallelism_type: Optional[torch.dtype] = torch.bfloat16,
    virtual_pipeline_parallelism: Optional[int] = None,
    context_parallelism: int = 1,
    sequence_parallelism: bool = False,
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    max_steps: int = 300000,
    precision: str = "bf16-mixed",
    accumulate_grad_batches: int = 1,
    gradient_clip_val: float = 1.0,
    limit_test_batches: int = 32,
    limit_val_batches: int = 32,
    log_every_n_steps: int = 10,
    val_check_interval: int = 1000,
    # Data
    global_batch_size=32,
    micro_batch_size=2,
    seq_length=4096,
    # Optimizer
    warmup_steps=500,
    constant_steps=0,
    min_lr=3e-5,
    max_lr=3e-4,
    # Training function
    fn=pretrain,
) -> run.Partial:
    """
    Create a pre-training recipe for Starcoder 15B model.

    This function sets up a complete configuration for pre-training, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the pre-training run.
        tensor_parallelism (int): Degree of tensor model parallelism.
        pipeline_parallelism (int): Degree of pipeline model parallelism.
        pipeline_parallelism_type (Optional[torch.dtype]): Data type for pipeline parallelism.
        virtual_pipeline_parallelism (Optional[int]): Size of virtual pipeline parallelism.
        context_parallelism (int): Degree of context parallelism.
        sequence_parallelism (bool): Whether to use sequence parallelism.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        max_steps (int): Maximum number of training steps.
        precision (str): Precision configuration, one of fp32, 16-mixed or bf16-mixed.
        accumulate_grad_batches (int): Number of steps per gradient accumulation.
        gradient_clip_val (float): Value for gradient clipping.
        limit_test_batches (int): Limit the number of test batches.
        limit_val_batches (int): Limit the number of validation batches.
        log_every_n_steps (int): Log every n steps.
        val_check_interval (int): Run validation every N steps.
        global_batch_size (int): Global batch size.
        micro_batch_size (int): Micro batch size.
        seq_length (int): Sequence length.
        warmup_steps (int): Number of warmup steps.
        constant_steps (int): Number of constant steps.
        min_lr (float): Minimum learning rate.
        max_lr (float): Maximum learning rate.
        fn (Callable): The pre-training function to use.

    Returns:
        run.Partial: Partial configuration for pre-training.

    Examples:
        CLI usage:
            $ nemo llm pretrain --factory starcoder_15b
            $ nemo llm pretrain --factory "starcoder_15b(num_nodes=1, name='my_starcoder2_pretrain')"

        Python API usage:
            >>> recipe = pretrain_recipe(name="starcoder2_pretrain", num_nodes=1)
            >>> print(recipe)

    Note:
        This recipe uses a mock dataset, look for the finetune examples to see how to change the dataset.
    """
    return run.Partial(
        fn,
        model=model(),
        trainer=starcoder_trainer(
            tensor_parallelism=tensor_parallelism,
            pipeline_parallelism=pipeline_parallelism,
            pipeline_parallelism_type=pipeline_parallelism_type,
            virtual_pipeline_parallelism=virtual_pipeline_parallelism,
            context_parallelism=context_parallelism,
            sequence_parallelism=sequence_parallelism,
            num_nodes=num_nodes,
            num_gpus_per_node=num_gpus_per_node,
            max_steps=max_steps,
            precision=precision,
            accumulate_grad_batches=accumulate_grad_batches,
            limit_test_batches=limit_test_batches,
            limit_val_batches=limit_val_batches,
            log_every_n_steps=log_every_n_steps,
            val_check_interval=val_check_interval,
            callbacks=[run.Config(TimingCallback)],
        ),
        data=run.Config(
            MockDataModule,
            seq_length=seq_length,
            global_batch_size=global_batch_size,
            micro_batch_size=micro_batch_size,
        ),
        log=default_log(dir=dir, name=name, tensorboard_logger=tensorboard_logger(name=name)),
        optim=distributed_fused_adam_with_cosine_annealing(
            precision=precision,
            warmup_steps=warmup_steps,
            constant_steps=constant_steps,
            min_lr=min_lr,
            max_lr=max_lr,
            clip_grad=gradient_clip_val,
        ),
        resume=default_resume(),
    )


@run.cli.factory(target=finetune, name=NAME)
def finetune_recipe(
    dir: Optional[str] = None,
    name: str = "default",
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    peft_scheme: Optional[str] = 'lora',
) -> run.Partial:
    """
    Create a fine-tuning recipe for Starcoder 15B model.

    This function sets up a complete configuration for fine-tuning, including
    model, trainer, data, logging, optimization, and resumption settings.
    The recipe uses LoRA (Low-Rank Adaptation) for efficient fine-tuning, unless peft_scheme is set to None.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the fine-tuning run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        peft_scheme (Optional[str]): Name of the peft scheme to use for fine-tuning.
            Allowed values: 'lora'/'dora'/'none'/None.

    Returns:
        run.Partial: Partial configuration for fine-tuning.

    Examples:
        CLI usage:
            $ nemo llm finetune --factory starcoder_15b

        Python API usage:
            >>> recipe = finetune_recipe(name="starcoder_15b_finetune", num_nodes=2)
            >>> print(recipe)

    Note:
        This recipe uses the SQuAD dataset for fine-tuning.
    """
    recipe = default_finetune_recipe(model(), "bigcode/starcoder", dir, name, num_nodes, num_gpus_per_node)
    if peft_scheme is None or peft_scheme.lower() == 'none':
        recipe.trainer.strategy.pipeline_model_parallel_size = 8
        recipe.optim.config.lr = 5e-6
    elif peft_scheme.lower() in ['lora', 'dora']:
        recipe.peft = run.Config(PEFT_STR2CLS[peft_scheme.lower()])
        recipe.optim.config.lr = 1e-4
    else:
        raise ValueError(f"Unrecognized peft scheme: {peft_scheme}")
    return recipe



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Callable, Optional

import lightning.pytorch as pl
import nemo_run as run

from nemo.collections.llm.api import finetune, pretrain
from nemo.collections.llm.gpt.data.mock import MockDataModule
from nemo.collections.llm.gpt.data.packed_sequence import PackedSequenceSpecs
from nemo.collections.llm.gpt.model.deepseek import DeepSeekModel, DeepSeekV2Config
from nemo.collections.llm.peft import PEFT_STR2CLS
from nemo.collections.llm.recipes.deepseek import trainer
from nemo.collections.llm.recipes.finetune_default import default_finetune_recipe
from nemo.collections.llm.recipes.log.default import default_log, default_resume, tensorboard_logger
from nemo.collections.llm.recipes.optim.adam import distributed_fused_adam_with_cosine_annealing
from nemo.lightning.pytorch.callbacks.deepep import DeepEPCallback
from nemo.utils.exp_manager import TimingCallback

NAME = "deepseek_v2"


@run.cli.factory(name=NAME)
def model() -> run.Config[pl.LightningModule]:
    """
    Factory function to create a DeepSeek-V2 (236B) model configuration.

    Returns:
        run.Config[pl.LightningModule]: Configuration for the DeepSeek V2 model.

    Examples:
        CLI usage:
            $ nemo llm pretrain model=deepseek_v2 ...

        Python API usage:
            >>> model_config = model()
            >>> print(model_config)
    """
    conf = run.Config(DeepSeekV2Config)
    return run.Config(DeepSeekModel, config=conf)


@run.cli.factory(target=pretrain, name=NAME)
def pretrain_recipe(
    dir: Optional[str] = None,
    name: str = "default",
    num_nodes: int = 16,
    num_gpus_per_node: int = 8,
    fn: Callable = pretrain,
) -> run.Partial:
    """
    Create a pre-training recipe for DeepSeek-V2 (236B) model.

    This function sets up a complete configuration for pre-training, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the pre-training run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        performance_mode (bool): If true, enables optimizations for maximum performance.
        fn (Callable): The pre-training function to use.

    Returns:
        run.Partial: Partial configuration for pre-training.

    Examples:
        CLI usage:
            $ nemo llm pretrain --factory deepseek_v2
            $ nemo llm pretrain --factory "deepseek_v2(num_nodes=16, name='my_deepseek_v2')"

        Python API usage:
            >>> recipe = pretrain_recipe(name="deepseek_v2_pretrain", num_nodes=16)
            >>> print(recipe)

    """
    recipe = run.Partial(
        fn,
        model=model(),
        trainer=trainer(
            tensor_parallelism=1,
            pipeline_parallelism=4,
            expert_parallelism=32,
            num_nodes=num_nodes,
            num_gpus_per_node=num_gpus_per_node,
            callbacks=[run.Config(TimingCallback)],
        ),
        data=run.Config(MockDataModule, seq_length=4096, global_batch_size=512, micro_batch_size=1),
        log=default_log(dir=dir, name=name, tensorboard_logger=tensorboard_logger(name=name)),
        optim=distributed_fused_adam_with_cosine_annealing(max_lr=3e-4),
        resume=default_resume(),
    )

    recipe.model.config.recompute_granularity = "full"
    recipe.model.config.recompute_method = "uniform"
    recipe.model.config.recompute_num_layers = 1

    # Use DeepEP
    deepep_callback = run.Config(DeepEPCallback)
    recipe.trainer.callbacks.append(deepep_callback)

    return recipe


@run.cli.factory(target=finetune, name=NAME)
def finetune_recipe(
    dir: Optional[str] = None,
    resume_path: str = "deepseek-ai/DeepSeek-V2",
    name: str = "default",
    num_nodes: int = 2,
    num_gpus_per_node: int = 8,
    peft_scheme: Optional[str] = 'lora',
    seq_length: Optional[int] = None,
    packed_sequence: Optional[bool] = None,
) -> run.Partial:
    """
    Create a fine-tuning recipe for DeepSeek-V2 (236B) model.

    This function sets up a complete configuration for fine-tuning, including
    model, trainer, data, logging, optimization, and resumption settings.
    The recipe uses LoRA (Low-Rank Adaptation) for efficient fine-tuning, unless peft_scheme is set to None.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        resume_path (str): Path to the NeMo checkpoint
        name (str): Name of the fine-tuning run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        peft_scheme (Optional[str]): Name of the peft scheme to use for fine-tuning.
            Allowed values: 'lora'/'dora'/'none'/None.
        seq_length (int): Maximum number of tokens per microbatch.
        packed_sequence (Optional[bool]): If true, fine-tuning sequences will be packed into batches up to the given
            maximum seq_length for better efficiency. By default, this value equals performance_mode.
    Returns:
        run.Partial: Partial configuration for fine-tuning.

    Examples:
        CLI usage:
            $ nemo llm finetune --factory deepseek_v2
            $ nemo llm finetune --factory "deepseek_v2(num_nodes=2, name='my_deepseek_v2_finetune')"

        Python API usage:
            >>> recipe = finetune_recipe(name="deepseek_v2_finetune", num_nodes=2)
            >>> print(recipe)

    Note:
        This recipe uses the SQuAD dataset for fine-tuning. Be aware that fine-tuning the DeepSeek-V2 model
        requires substantial computational resources.
    """

    if seq_length is None:
        seq_length = 2048

    if num_nodes is None:
        if peft_scheme is None or peft_scheme.lower() == 'none':
            num_nodes = 16
        elif peft_scheme.lower() in ['lora', 'dora']:
            num_nodes = 2

    recipe = default_finetune_recipe(model(), resume_path, dir, name, num_nodes, num_gpus_per_node, packed_sequence)
    if peft_scheme is None or peft_scheme.lower() == 'none':
        recipe.trainer.strategy.pipeline_model_parallel_size = 4
        recipe.trainer.strategy.expert_model_parallel_size = 32
        recipe.optim.config.lr = 5e-6
    elif peft_scheme.lower() in ['lora', 'dora']:
        recipe.peft = run.Config(PEFT_STR2CLS[peft_scheme.lower()])
        recipe.peft.target_modules = [
            'linear_q_down_proj',
            'linear_q_up_proj',
            'linear_kv_down_proj',
            'linear_kv_up_proj',
            'linear_proj',
        ]
        recipe.optim.config.use_distributed_optimizer = False
        recipe.model.config.cross_entropy_loss_fusion = False
        recipe.trainer.strategy.tensor_model_parallel_size = 1
        recipe.trainer.strategy.pipeline_model_parallel_size = 1
        recipe.trainer.strategy.expert_model_parallel_size = 16
        recipe.optim.config.lr = 1e-4
    else:
        raise ValueError(f"Unrecognized peft scheme: {peft_scheme}")

    # Sequence length settings in the model and dataset must agree
    recipe.model.config.seq_length = seq_length
    recipe.data.seq_length = seq_length
    if packed_sequence:
        recipe.data.dataset_kwargs = {'pad_to_max_length': True}
        recipe.data.packed_sequence_specs = run.Config(PackedSequenceSpecs, packed_sequence_size=seq_length)

    return recipe



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import csv
import os
import re
import string
from typing import Dict, List, Union

__all__ = [
    "read_wordids",
    "set_grapheme_case",
    "GRAPHEME_CASE_UPPER",
    "GRAPHEME_CASE_LOWER",
    "GRAPHEME_CASE_MIXED",
    "get_heteronym_spans",
]


# define grapheme cases.
GRAPHEME_CASE_UPPER = "upper"
GRAPHEME_CASE_LOWER = "lower"
GRAPHEME_CASE_MIXED = "mixed"


def read_wordids(wordid_map: str):
    """
    Reads wordid file from WikiHomograph dataset,
    https://github.com/google-research-datasets/WikipediaHomographData/blob/master/data/wordids.tsv

    Args:
        wordid_map: path to wordids.tsv
    Returns:
        data_dict: a dictionary of graphemes with corresponding word_id - ipa_form pairs
        wordid_to_idx: word id to label id mapping
    """
    if not os.path.exists(wordid_map):
        raise ValueError(f"{wordid_map} not found")

    data_dict = {}
    wordid_to_idx = {}

    with open(wordid_map, "r", encoding="utf-8") as f:
        tsv_file = csv.reader(f, delimiter="\t")

        for i, line in enumerate(tsv_file):
            if i == 0:
                continue
            grapheme = line[0]
            word_id = line[1]

            ipa_form = line[3]
            wordid_to_idx[word_id] = len(wordid_to_idx)
            if grapheme not in data_dict:
                data_dict[grapheme] = {}
            data_dict[grapheme][word_id] = ipa_form
    return data_dict, wordid_to_idx


def get_wordid_to_phonemes(wordid_to_phonemes_file: str, to_lower: bool = True):
    """
    WikiHomograph and NeMo use slightly different phoneme sets, this function reads WikiHomograph word_ids to NeMo
    IPA heteronyms mapping.

    Args:
        wordid_to_phonemes_file: Path to a file with mapping from wordid predicted by the model to phonemes, e.g.,
            NeMo/scripts/tts_dataset_files/wordid_to_ipa-0.7b_nv22.10.tsv
        to_lower: set to True to lower case wordid
    """
    if not os.path.exists(wordid_to_phonemes_file):
        raise ValueError(f"{wordid_to_phonemes_file} not found")

    wordid_to_nemo_cmu = {}
    with open(wordid_to_phonemes_file, "r", encoding="utf-8") as f:
        for i, line in enumerate(f):
            if to_lower:
                line = line.lower()
            line = line.strip().split("  ")
            wordid_to_nemo_cmu[line[0]] = line[1]
    return wordid_to_nemo_cmu


def remove_punctuation(text: str, remove_spaces: bool = False, do_lower: bool = False, exclude: List[str] = None):
    """
    Remove punctuation marks form text

    Args:
        text: input text
        remove_spaces: set to True to remove spaces
        do_lower: set to True to lower case the text
        exclude: specify list of punctuation marks keep in the output, e.g., exclude=["'", "."]

    Return:
        processed text with punctuation marks removed
    """
    all_punct_marks = string.punctuation

    if exclude is not None:
        for p in exclude:
            all_punct_marks = all_punct_marks.replace(p, "")

    text = re.sub("[" + all_punct_marks + "]", " ", text)

    text = re.sub(r" +", " ", text)
    if remove_spaces:
        text = text.replace(" ", "").replace("\u00A0", "").strip()

    if do_lower:
        text = text.lower()
    return text.strip()


def get_heteronym_spans(sentences: List[str], supported_heteronyms: Union[Dict, List]):
    """
    Find heteronyms in sentences and returns span indices

    Args:
        sentences: sentences to find heteronyms in
        supported_heteronyms: heteronyms to look for

    Return:
        start_end: List[Tuple[int]] - start-end indices that indicate location of found heteronym in the sentence
        heteronyms: List[List[str]] - heteronyms found in sentences, each sentence can contain more than one heteronym
    """
    start_end = []
    heteronyms = []
    for sent in sentences:
        cur_start_end = []
        cur_heteronyms = []
        start_idx = 0
        for word in sent.lower().split():
            word_by_hyphen = word.split("-")
            for sub_word in word_by_hyphen:
                no_punct_word = remove_punctuation(sub_word, do_lower=True, remove_spaces=False)
                if no_punct_word in supported_heteronyms:
                    start_idx = sent.lower().index(no_punct_word, start_idx)
                    end_idx = start_idx + len(no_punct_word)
                    cur_start_end.append((start_idx, end_idx))
                    cur_heteronyms.append(no_punct_word)
                    start_idx = end_idx
                else:
                    start_idx += len(sub_word) + 1
        heteronyms.append(cur_heteronyms)
        start_end.append(cur_start_end)
    return start_end, heteronyms


def set_grapheme_case(text: str, case: str = "upper") -> str:
    if case == "upper":
        text_new = text.upper()
    elif case == "lower":
        text_new = text.lower()
    elif case == "mixed":  # keep as-is, mix-cases
        text_new = text
    else:
        raise ValueError(f"Case <{case}> is not supported. Please specify either 'upper', 'lower', or 'mixed'.")

    return text_new



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2020, Xiaomi CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple

import torch

from nemo.collections.asr.parts.k2.utils import add_self_loops, compose_with_self_loops, intersect_with_self_loops

from nemo.core.utils.k2_guard import k2  # import k2 from guard module


class CtcTopologyCompiler(object):
    """Default graph compiler.
    It applies its topology to the input token sequence to compile the supervision graph.
    
    Based on https://github.com/k2-fsa/snowfall/blob/master/snowfall/training/ctc_graph.py
    """

    def __init__(
        self,
        num_classes: int,
        blank: int,
        topo_type: str = "default",
        topo_with_self_loops: bool = True,
        device: torch.device = torch.device("cpu"),
    ):
        self.topo_type = topo_type
        self.device = device
        from nemo.collections.asr.parts.k2.topologies import build_topo

        self.base_graph = k2.arc_sort(build_topo(topo_type, list(range(num_classes)), blank, topo_with_self_loops)).to(
            self.device
        )
        self.ctc_topo_inv = k2.arc_sort(self.base_graph.invert())

    def to(self, device: torch.device):
        self.ctc_topo_inv = self.ctc_topo_inv.to(device)
        if self.base_graph is not None:
            self.base_graph = self.base_graph.to(device)
        self.device = device

    def compile(self, targets: torch.Tensor, target_lengths: torch.Tensor) -> 'k2.Fsa':
        token_ids_list = [t[:l].tolist() for t, l in zip(targets, target_lengths)]
        label_graph = k2.linear_fsa(token_ids_list).to(self.device)
        label_graph.aux_labels = label_graph.labels.clone()
        supervision_graphs = compose_with_self_loops(self.base_graph, label_graph)
        supervision_graphs = k2.arc_sort(supervision_graphs).to(self.device)

        # make sure the gradient is not accumulated
        supervision_graphs.requires_grad_(False)
        return supervision_graphs


class CtcNumGraphCompiler(CtcTopologyCompiler):
    """Graph compiler with auxiliary graph to compose with the topology.
    The supervision graph contains the auxiliary graph information.
    """

    def __init__(
        self,
        num_classes: int,
        blank: int,
        topo_type: str = "default",
        topo_with_self_loops: bool = True,
        device: torch.device = torch.device("cpu"),
        aux_graph: Optional['k2.Fsa'] = None,
    ):
        super().__init__(num_classes, blank, topo_type, topo_with_self_loops, device)
        if aux_graph is None:
            self.decoding_graph = k2.create_fsa_vec([self.ctc_topo_inv.invert()]).to(self.device)
        else:
            self.base_graph = intersect_with_self_loops(self.ctc_topo_inv, aux_graph).invert_()
            self.base_graph = k2.arc_sort(self.base_graph).to(self.device)

    def compile(
        self, targets: torch.Tensor, target_lengths: torch.Tensor, aux_graph: Optional['k2.Fsa'] = None,
    ) -> 'k2.Fsa':
        if aux_graph is None and self.base_graph is None:
            raise ValueError(
                f"At least one of aux_graph and self.base_graph must be set: {aux_graph}, {self.base_graph}"
            )
        elif aux_graph is not None:
            self.base_graph = intersect_with_self_loops(self.ctc_topo_inv, aux_graph).invert()
            self.base_graph = k2.arc_sort(self.base_graph).to(self.device)
        return super().compile(targets, target_lengths)


class MmiGraphCompiler(CtcNumGraphCompiler):
    """Graph compiler for MMI loss.
    The decoding graph is a composition of the auxiliary graph and the topology.
    It is returned along with the supervision graph on every compile() call.
    """

    def __init__(
        self,
        num_classes: int,
        blank: int,
        topo_type: str = "default",
        topo_with_self_loops: bool = True,
        device: torch.device = torch.device("cpu"),
        aux_graph: Optional['k2.Fsa'] = None,
    ):
        super().__init__(num_classes, blank, topo_type, topo_with_self_loops, device, aux_graph)
        if aux_graph is None:
            self.decoding_graph = k2.create_fsa_vec([self.ctc_topo_inv.invert()]).to(self.device)
        else:
            self.decoding_graph = k2.create_fsa_vec([self.base_graph.detach()]).to(self.device)

    def to(self, device: torch.device):
        if self.decoding_graph is not None:
            self.decoding_graph = self.decoding_graph.to(device)
        super().to(device)

    def compile(
        self, targets: torch.Tensor, target_lengths: torch.Tensor, aux_graph: Optional['k2.Fsa'] = None,
    ) -> Tuple['k2.Fsa', 'k2.Fsa']:
        supervision_graphs = super().compile(targets, target_lengths, aux_graph)
        if aux_graph is None and self.decoding_graph is None:
            raise ValueError(
                f"At least one of aux_graph and self.decoding_graph must be set: {aux_graph}, {self.decoding_graph}"
            )
        elif aux_graph is not None:
            self.decoding_graph = k2.create_fsa_vec([self.base_graph.detach()]).to(self.device)
        return supervision_graphs, self.decoding_graph


class RnntTopologyCompiler(CtcTopologyCompiler):
    """Default graph compiler for RNNT loss.
    Each supervision graph is composed with the corresponding RNNT emission adapter.

    If max_adapter_length is provided, the maximum adapter length is limited.

    Note:
      The actual number of classes is `num_classes` + 1 with <eps> as the class 0.

    Warning:
      It is currently not recommended to use topologies other than "minimal".
    """

    def __init__(
        self,
        num_classes: int,
        blank: int,
        topo_type: str = "minimal",
        topo_with_self_loops: bool = True,
        device: torch.device = torch.device("cpu"),
        max_adapter_length: int = 0,
    ):
        if topo_type == "compact":
            raise NotImplementedError(f"This compiler does not support topo_type==`compact`.")
        super().__init__(num_classes, blank, topo_type, topo_with_self_loops, device)
        from nemo.collections.asr.parts.k2.topologies import RnntEmissionAdapterBuilder

        self.max_adapter_length = max_adapter_length
        self._builder = RnntEmissionAdapterBuilder(list(range(num_classes)), blank, num_classes)

    def compile(self, targets: torch.Tensor, target_lengths: torch.Tensor) -> 'k2.Fsa':
        supervision_graphs = add_self_loops(super().compile(targets, target_lengths), self._builder.eps_num, "input")

        adapters = self._builder(
            torch.where(target_lengths > self.max_adapter_length, self.max_adapter_length, target_lengths)
            if self.max_adapter_length > 0 and self.max_adapter_length < target_lengths.max()
            else target_lengths
        ).to(device=self.device)
        return k2.intersect(adapters, supervision_graphs, treat_epsilons_specially=False)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Tuple

from omegaconf import DictConfig, open_dict

from nemo.core.classes.mixins.adapter_mixins import AdapterModelPTMixin, AdapterModuleMixin
from nemo.utils import logging, logging_mode


class ASRAdapterModelMixin(AdapterModelPTMixin):
    """ASR Adapter Mixin that can augment any Encoder module with Adapter module support.

    This mixin class should be used only with a top level ModelPT subclass, that includes an `encoder` submodule.
    This mixin class adds several utility methods which are propagated to the `encoder`.

    An Adapter module is any Pytorch nn.Module that possess a few properties :

    - It's input and output dimension are the same, while the hidden dimension need not be the same.
    - The final layer of the Adapter module is zero-initialized, so that the residual connection to the adapter
        yields the original output.

    This mixin adds the following instance variables to the class this inherits it:

        -   `adapter_layer`: A torch.nn.ModuleDict(), whose keys are the names of the adapter (globally unique),
                and values are the Adapter nn.Module().
        -   `adapter_cfg`: A OmegaConf DictConfig object that holds the config of the adapters that are initialized.
        -   `adapter_global_cfg_key`: A str representing a key in the model config that can be provided by the user.
                The value resolves to `global_cfg`, and can be overridden via `model.cfg.adapters.global_cfg.*`.

    **Note**: This module **is** responsible for maintaining its config. At the ModelPT level, it will access and
        write Adapter config information to `self.cfg.adapters`.
    """

    def setup_adapters(self):
        """
        Utility method that is called in the ASR ModelPT-implementation constructor, so as to restore any
        adapters that were previously added.

        This method should be called just once at constructor time.
        """
        supports_adapters = False

        # At least the encoder must extend AdapterModuleMixin
        valid_adapter_names = [x for x in self.adapter_module_names if x != '']
        for module_name in valid_adapter_names:
            if hasattr(self, module_name) and isinstance(getattr(self, module_name), AdapterModuleMixin):
                supports_adapters |= True

        # If adapters are supported, setup the adapter config + any modules (pre-existing adapter modules)
        if supports_adapters:
            super().setup_adapters()

    def add_adapter(self, name: str, cfg: DictConfig):
        """
        Add an Adapter module to this model.

        Args:
            name: A globally unique name for the adapter. Will be used to access, enable and disable adapters.
            cfg: A DictConfig that contains at the bare minimum `__target__` to instantiate a new Adapter module.
        """
        # setup the config for adapters
        super().add_adapter(name=name, cfg=cfg)

        # Resolve module name and adapter name
        module_name, _ = self.resolve_adapter_module_name_(name)

        # Use + as a splitter, in order to share one name across multiple modules
        if '+' in module_name:
            module_names = module_name.split('+')
        else:
            module_names = [module_name]

        valid_module_names = [x for x in self.adapter_module_names if x != '']
        default_module_name = self.default_adapter_module_name

        # Check if default module name is None or not
        if default_module_name is None:
            raise ValueError(
                f"Default module name is None. Class {self.__class__.__name__} must implement "
                f"`default_adapter_module_name`"
            )

        # Update the model.cfg with information about the new adapter from cfg
        with open_dict(self.cfg):
            for module_name in module_names:
                # Check if encoder adapters should be added
                if module_name == '':
                    if hasattr(self, default_module_name):
                        # Dispatch the call to the default model.
                        getattr(self, default_module_name).add_adapter(name=name, cfg=cfg)

                elif module_name in valid_module_names:
                    # Check if module exists
                    if hasattr(self, module_name):
                        # Dispatch the call to the module.
                        getattr(self, module_name).add_adapter(name=name, cfg=cfg)

    def is_adapter_available(self) -> bool:
        """
        Checks if any Adapter module has been instantiated.

        Returns:
            bool, determining if any Adapter module has been instantiated. Returns true even if the adapters are
            enabled or disabled, false only if no adapters exist.
        """
        config_contains_adapter = super().is_adapter_available()

        valid_module_names = [x for x in self.adapter_module_names if x != '']

        # Forward the method call to the individual modules
        for module_name in valid_module_names:
            if hasattr(self, module_name) and isinstance(getattr(self, module_name), AdapterModuleMixin):
                config_contains_adapter |= getattr(self, module_name).is_adapter_available()

        return config_contains_adapter

    def set_enabled_adapters(self, name: Optional[str] = None, enabled: bool = True):
        """
        Updated the internal adapter config, determining if an adapter (or all adapters) are either
        enabled or disabled.

        A common user pattern would be to disable all adapters (either after adding them, or restoring a model
        with pre-existing adapters) and then simply enable one of the adapters.

        .. code::

            model.set_enabled_adapters(enabled=False)
            model.set_enabled_adapters(name=<some adapter name>, enabled=True)

        Args:
            name: Optional str. If a str name is given, the config will be updated to the value of `enabled`.
                If no name is given, then all adapters will be enabled/disabled.
            enabled: Bool, determines if the adapter(s) will be enabled/disabled.
        """
        super().set_enabled_adapters(name=name, enabled=enabled)

        # Resolve the module name and adapter name
        if name is not None:
            module_name, _ = self.resolve_adapter_module_name_(name)
        else:
            module_name = None

        # Use + as a splitter, in order to share one name across multiple modules
        if module_name is not None and '+' in module_name:
            module_names = module_name.split('+')
        else:
            module_names = [module_name]

        valid_module_names = [x for x in self.adapter_module_names if x != '']
        default_module_name = self.default_adapter_module_name

        # Check if default module name is None or not
        if default_module_name is None:
            raise ValueError(
                f"Default module name is None. Class {self.__class__.__name__} must implement "
                f"`default_adapter_module_name`"
            )

        # Forward the method call to the individual modules if they exist
        for module_name in module_names:
            # Check if encoder adapters should be used

            if module_name == '':
                if hasattr(self, default_module_name):
                    # Dispatch the call to the default model.
                    getattr(self, default_module_name).set_enabled_adapters(name=name, enabled=enabled)

            elif module_name in valid_module_names:
                if hasattr(self, module_name):
                    # Dispatch the call to the module.
                    getattr(self, module_name).set_enabled_adapters(name=name, enabled=enabled)

    def get_enabled_adapters(self) -> List[str]:
        """
        Returns a list of all enabled adapters.

        Returns:
            A list of str names of each enabled adapter(s).
        """
        enabled_adapters = super().get_enabled_adapters()

        valid_module_names = [x for x in self.adapter_module_names if x != '']

        # Check if encoder adapters should be used or are enabled
        for module_name in valid_module_names:
            if hasattr(self, module_name) and isinstance(getattr(self, module_name), AdapterModuleMixin):
                enabled_adapters.extend(getattr(self, module_name).get_enabled_adapters())

        enabled_adapters = list(sorted(list(set(enabled_adapters))))

        return enabled_adapters

    def check_valid_model_with_adapter_support_(self):
        """
        Utility method to test if the subclass of this mixin is an appropriate subclass of ModelPT itself.
        """
        # Obtain the global adapter config if possible, otherwise use sensible defaults.
        global_cfg = self._get_global_cfg()

        valid_module_names = [x for x in self.adapter_module_names if x != '']

        for module_name in valid_module_names:
            check_adapter_support = global_cfg.get(f'check_{module_name}_adapter', True)

            if check_adapter_support:
                # Test whether the module supports adapters
                if hasattr(self, module_name) and not isinstance(getattr(self, module_name), AdapterModuleMixin):
                    logging.warning(
                        f'Module `{module_name}` exists, but {getattr(self, module_name).__class__.__name__} '
                        f'does not implement `AdapterModuleMixin`',
                        mode=logging_mode.ONCE,
                    )

    def resolve_adapter_module_name_(self, name: str) -> Tuple[str, str]:
        """
        Utility method to resolve a given global/module adapter name to its components.
        Always returns a tuple representing (module_name, adapter_name). ":" is used as the
        delimiter for denoting the module name vs the adapter name.

        Will attempt to also resolve a given adapter_name alone back to (module_name, adapter_name)
        if the metadata config exists for access.

        Args:
            name: A global adapter, or a module adapter name (with structure module_name:adapter_name).

        Returns:
            A tuple representing (module_name, adapter_name). If a global adapter is provided,
            module_name is set to ''.
        """
        module_name, adapter_name = super().resolve_adapter_module_name_(name)

        # Use + as a splitter, in order to share one name across multiple modules
        if '+' in module_name:
            module_names = module_name.split('+')
        else:
            module_names = [module_name]

        # resolve name and module only for valid modules
        valid_module_names = self.adapter_module_names

        for mod_name in module_names:
            if mod_name not in valid_module_names:
                raise ValueError(f"Provided module name `{mod_name}` is not in valid list : {valid_module_names}")

        return (module_name, adapter_name)

    def _get_global_cfg(self):
        """
        Utility method, to either extract or construct the global config inside adapters config.
        """
        global_config = DictConfig({})
        if 'adapters' in self.cfg and self.adapter_global_cfg_key in self.cfg.adapters:
            global_config = self.adapter_cfg[self.adapter_global_cfg_key]
        return global_config

    @property
    def adapter_module_names(self) -> List[str]:
        valid_module_names = ['', 'encoder', 'decoder', 'joint']
        return valid_module_names

    @property
    def default_adapter_module_name(self) -> str:
        return 'encoder'



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
from typing import Dict, List, Optional, Union

import torch
from lightning.pytorch import Trainer
from omegaconf import DictConfig, ListConfig, OmegaConf, open_dict

from nemo.collections.asr.data import audio_to_text_dataset
from nemo.collections.asr.data.audio_to_text import _AudioTextDataset
from nemo.collections.asr.data.audio_to_text_dali import AudioToBPEDALIDataset
from nemo.collections.asr.data.audio_to_text_lhotse import LhotseSpeechToTextBpeDataset
from nemo.collections.asr.losses.ctc import CTCLoss
from nemo.collections.asr.losses.rnnt import RNNTLoss
from nemo.collections.asr.metrics.wer import WER
from nemo.collections.asr.models.hybrid_rnnt_ctc_models import EncDecHybridRNNTCTCModel
from nemo.collections.asr.parts.mixins import ASRBPEMixin
from nemo.collections.asr.parts.submodules.ctc_decoding import CTCBPEDecoding, CTCBPEDecodingConfig
from nemo.collections.asr.parts.submodules.rnnt_decoding import RNNTBPEDecoding, RNNTBPEDecodingConfig
from nemo.collections.asr.parts.utils.asr_batching import get_semi_sorted_batch_sampler
from nemo.collections.common.data.lhotse import get_lhotse_dataloader_from_config
from nemo.core.classes.common import PretrainedModelInfo
from nemo.utils import logging, model_utils


class EncDecHybridRNNTCTCBPEModel(EncDecHybridRNNTCTCModel, ASRBPEMixin):
    """Base class for encoder decoder RNNT-based models with auxiliary CTC decoder/loss and subword tokenization."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        # Convert to Hydra 1.0 compatible DictConfig
        cfg = model_utils.convert_model_config_to_dict_config(cfg)
        cfg = model_utils.maybe_update_config_version(cfg, make_copy=False)

        # Tokenizer is necessary for this model
        if 'tokenizer' not in cfg:
            raise ValueError("`cfg` must have `tokenizer` config to create a tokenizer !")

        if not isinstance(cfg, DictConfig):
            cfg = OmegaConf.create(cfg)

        # Setup the tokenizer
        self._setup_tokenizer(cfg.tokenizer)

        # Initialize a dummy vocabulary
        vocabulary = self.tokenizer.tokenizer.get_vocab()

        # Set the new vocabulary
        with open_dict(cfg):
            cfg.labels = ListConfig(list(vocabulary))

        with open_dict(cfg.decoder):
            cfg.decoder.vocab_size = len(vocabulary)

        with open_dict(cfg.joint):
            cfg.joint.num_classes = len(vocabulary)
            cfg.joint.vocabulary = ListConfig(list(vocabulary))
            cfg.joint.jointnet.encoder_hidden = cfg.model_defaults.enc_hidden
            cfg.joint.jointnet.pred_hidden = cfg.model_defaults.pred_hidden

        # setup auxiliary CTC decoder
        if 'aux_ctc' not in cfg:
            raise ValueError(
                "The config need to have a section for the CTC decoder named as aux_ctc for Hybrid models."
            )

        with open_dict(cfg):
            if self.tokenizer_type == "agg":
                cfg.aux_ctc.decoder.vocabulary = ListConfig(vocabulary)
            else:
                cfg.aux_ctc.decoder.vocabulary = ListConfig(list(vocabulary.keys()))

        if cfg.aux_ctc.decoder["num_classes"] < 1:
            logging.info(
                "\nReplacing placholder number of classes ({}) with actual number of classes - {}".format(
                    cfg.aux_ctc.decoder["num_classes"], len(vocabulary)
                )
            )
            cfg.aux_ctc.decoder["num_classes"] = len(vocabulary)

        super().__init__(cfg=cfg, trainer=trainer)

        self.cfg.decoding = self.set_decoding_type_according_to_loss(self.cfg.decoding)
        # Setup decoding object
        self.decoding = RNNTBPEDecoding(
            decoding_cfg=self.cfg.decoding,
            decoder=self.decoder,
            joint=self.joint,
            tokenizer=self.tokenizer,
        )

        # Setup wer object
        self.wer = WER(
            decoding=self.decoding,
            batch_dim_index=0,
            use_cer=self.cfg.get('use_cer', False),
            log_prediction=self.cfg.get('log_prediction', True),
            dist_sync_on_step=True,
        )

        # Setup fused Joint step if flag is set
        if self.joint.fuse_loss_wer:
            self.joint.set_loss(self.loss)
            self.joint.set_wer(self.wer)

        # Setup CTC decoding
        ctc_decoding_cfg = self.cfg.aux_ctc.get('decoding', None)
        if ctc_decoding_cfg is None:
            ctc_decoding_cfg = OmegaConf.structured(CTCBPEDecodingConfig)
            with open_dict(self.cfg.aux_ctc):
                self.cfg.aux_ctc.decoding = ctc_decoding_cfg
        self.ctc_decoding = CTCBPEDecoding(self.cfg.aux_ctc.decoding, tokenizer=self.tokenizer)

        # Setup CTC WER
        self.ctc_wer = WER(
            decoding=self.ctc_decoding,
            use_cer=self.cfg.aux_ctc.get('use_cer', False),
            dist_sync_on_step=True,
            log_prediction=self.cfg.get("log_prediction", False),
        )

        # setting the RNNT decoder as the default one
        self.cur_decoder = "rnnt"

    def _setup_dataloader_from_config(self, config: Optional[Dict]):

        if config.get("use_lhotse"):
            return get_lhotse_dataloader_from_config(
                config,
                # During transcription, the model is initially loaded on the CPU.
                # To ensure the correct global_rank and world_size are set,
                # these values must be passed from the configuration.
                global_rank=self.global_rank if not config.get("do_transcribe", False) else config.get("global_rank"),
                world_size=self.world_size if not config.get("do_transcribe", False) else config.get("world_size"),
                dataset=LhotseSpeechToTextBpeDataset(
                    tokenizer=self.tokenizer,
                    return_cuts=config.get("do_transcribe", False),
                ),
                tokenizer=self.tokenizer,
            )

        dataset = audio_to_text_dataset.get_audio_to_text_bpe_dataset_from_config(
            config=config,
            local_rank=self.local_rank,
            global_rank=self.global_rank,
            world_size=self.world_size,
            tokenizer=self.tokenizer,
            preprocessor_cfg=self.cfg.get("preprocessor", None),
        )

        if dataset is None:
            return None

        if isinstance(dataset, AudioToBPEDALIDataset):
            # DALI Dataset implements dataloader interface
            return dataset

        shuffle = config['shuffle']
        if isinstance(dataset, torch.utils.data.IterableDataset):
            shuffle = False

        if hasattr(dataset, 'collate_fn'):
            collate_fn = dataset.collate_fn
        elif hasattr(dataset.datasets[0], 'collate_fn'):
            # support datasets that are lists of entries
            collate_fn = dataset.datasets[0].collate_fn
        else:
            # support datasets that are lists of lists
            collate_fn = dataset.datasets[0].datasets[0].collate_fn

        batch_sampler = None
        if config.get('use_semi_sorted_batching', False):
            if not isinstance(dataset, _AudioTextDataset):
                raise RuntimeError(
                    "Semi Sorted Batch sampler can be used with AudioToCharDataset or AudioToBPEDataset "
                    f"but found dataset of type {type(dataset)}"
                )
            # set batch_size and batch_sampler to None to disable automatic batching
            batch_sampler = get_semi_sorted_batch_sampler(self, dataset, config)
            config['batch_size'] = None
            config['drop_last'] = False
            shuffle = False

        return torch.utils.data.DataLoader(
            dataset=dataset,
            batch_size=config['batch_size'],
            sampler=batch_sampler,
            batch_sampler=None,
            collate_fn=collate_fn,
            drop_last=config.get('drop_last', False),
            shuffle=shuffle,
            num_workers=config.get('num_workers', 0),
            pin_memory=config.get('pin_memory', False),
        )

    def _setup_transcribe_dataloader(self, config: Dict) -> 'torch.utils.data.DataLoader':
        """
        Setup function for a temporary data loader which wraps the provided audio file.

        Args:
            config: A python dictionary which contains the following keys:
            paths2audio_files: (a list) of paths to audio files. The files should be relatively short fragments. \
                Recommended length per file is between 5 and 25 seconds.
            batch_size: (int) batch size to use during inference. \
                Bigger will result in better throughput performance but would use more memory.
            temp_dir: (str) A temporary directory where the audio manifest is temporarily
                stored.
            num_workers: (int) number of workers. Depends of the batch_size and machine. \
                0 - only the main process will load batches, 1 - one worker (not main process)

        Returns:
            A pytorch DataLoader for the given audio file(s).
        """

        if 'manifest_filepath' in config:
            manifest_filepath = config['manifest_filepath']
            batch_size = config['batch_size']
        else:
            manifest_filepath = os.path.join(config['temp_dir'], 'manifest.json')
            batch_size = min(config['batch_size'], len(config['paths2audio_files']))

        dl_config = {
            'use_lhotse': config.get('use_lhotse', True),
            'manifest_filepath': manifest_filepath,
            'sample_rate': self.preprocessor._sample_rate,
            'batch_size': batch_size,
            'shuffle': False,
            'num_workers': config.get('num_workers', min(batch_size, os.cpu_count() - 1)),
            'pin_memory': True,
            'channel_selector': config.get('channel_selector', None),
            'use_start_end_token': self.cfg.validation_ds.get('use_start_end_token', False),
        }

        if config.get("augmentor"):
            dl_config['augmentor'] = config.get("augmentor")

        temporary_datalayer = self._setup_dataloader_from_config(config=DictConfig(dl_config))
        return temporary_datalayer

    def change_vocabulary(
        self,
        new_tokenizer_dir: Union[str, DictConfig],
        new_tokenizer_type: str,
        decoding_cfg: Optional[DictConfig] = None,
        ctc_decoding_cfg: Optional[DictConfig] = None,
    ):
        """
        Changes vocabulary used during RNNT decoding process. Use this method when fine-tuning on
        from pre-trained model. This method changes only decoder and leaves encoder and pre-processing
        modules unchanged. For example, you would use it if you want to use pretrained encoder when
        fine-tuning on data in another language, or when you'd need model to learn capitalization,
        punctuation and/or special characters.

        Args:
            new_tokenizer_dir: Directory path to tokenizer or a config for a new tokenizer (if the tokenizer type is `agg`)
            new_tokenizer_type: Type of tokenizer. Can be either `agg`, `bpe` or `wpe`.
            decoding_cfg: A config for the decoder, which is optional. If the decoding type
                needs to be changed (from say Greedy to Beam decoding etc), the config can be passed here.
            ctc_decoding_cfg: A config for auxiliary CTC decoding, which is optional and can be used to change the decoding type.

        Returns: None

        """
        if isinstance(new_tokenizer_dir, DictConfig):
            if new_tokenizer_type == 'agg':
                new_tokenizer_cfg = new_tokenizer_dir
            else:
                raise ValueError(
                    f'New tokenizer dir should be a string unless the tokenizer is `agg`, but this tokenizer type is: {new_tokenizer_type}'
                )
        else:
            new_tokenizer_cfg = None

        if new_tokenizer_cfg is not None:
            tokenizer_cfg = new_tokenizer_cfg
        else:
            if not os.path.isdir(new_tokenizer_dir):
                raise NotADirectoryError(
                    f'New tokenizer dir must be non-empty path to a directory. But I got: {new_tokenizer_dir}'
                )

            if new_tokenizer_type.lower() not in ('bpe', 'wpe'):
                raise ValueError(f'New tokenizer type must be either `bpe` or `wpe`, got {new_tokenizer_type}')

            tokenizer_cfg = OmegaConf.create({'dir': new_tokenizer_dir, 'type': new_tokenizer_type})

        # Setup the tokenizer
        self._setup_tokenizer(tokenizer_cfg)

        # Initialize a dummy vocabulary
        vocabulary = self.tokenizer.tokenizer.get_vocab()

        joint_config = self.joint.to_config_dict()
        new_joint_config = copy.deepcopy(joint_config)
        if self.tokenizer_type == "agg":
            new_joint_config["vocabulary"] = ListConfig(vocabulary)
        else:
            new_joint_config["vocabulary"] = ListConfig(list(vocabulary.keys()))

        new_joint_config['num_classes'] = len(vocabulary)
        del self.joint
        self.joint = EncDecHybridRNNTCTCBPEModel.from_config_dict(new_joint_config)

        decoder_config = self.decoder.to_config_dict()
        new_decoder_config = copy.deepcopy(decoder_config)
        new_decoder_config.vocab_size = len(vocabulary)
        del self.decoder
        self.decoder = EncDecHybridRNNTCTCBPEModel.from_config_dict(new_decoder_config)

        del self.loss
        self.loss = RNNTLoss(num_classes=self.joint.num_classes_with_blank - 1)

        if decoding_cfg is None:
            # Assume same decoding config as before
            decoding_cfg = self.cfg.decoding

        # Assert the decoding config with all hyper parameters
        decoding_cls = OmegaConf.structured(RNNTBPEDecodingConfig)
        decoding_cls = OmegaConf.create(OmegaConf.to_container(decoding_cls))
        decoding_cfg = OmegaConf.merge(decoding_cls, decoding_cfg)
        decoding_cfg = self.set_decoding_type_according_to_loss(decoding_cfg)

        self.decoding = RNNTBPEDecoding(
            decoding_cfg=decoding_cfg,
            decoder=self.decoder,
            joint=self.joint,
            tokenizer=self.tokenizer,
        )

        self.wer = WER(
            decoding=self.decoding,
            batch_dim_index=self.wer.batch_dim_index,
            use_cer=self.wer.use_cer,
            log_prediction=self.wer.log_prediction,
            dist_sync_on_step=True,
        )

        # Setup fused Joint step
        if self.joint.fuse_loss_wer or (
            self.decoding.joint_fused_batch_size is not None and self.decoding.joint_fused_batch_size > 0
        ):
            self.joint.set_loss(self.loss)
            self.joint.set_wer(self.wer)

        # Update config
        with open_dict(self.cfg.joint):
            self.cfg.joint = new_joint_config

        with open_dict(self.cfg.decoder):
            self.cfg.decoder = new_decoder_config

        with open_dict(self.cfg.decoding):
            self.cfg.decoding = decoding_cfg

        logging.info(f"Changed tokenizer of the RNNT decoder to {self.joint.vocabulary} vocabulary.")

        # set up the new tokenizer for the CTC decoder
        if hasattr(self, 'ctc_decoder'):
            ctc_decoder_config = copy.deepcopy(self.ctc_decoder.to_config_dict())
            # sidestepping the potential overlapping tokens issue in aggregate tokenizers
            if self.tokenizer_type == "agg":
                ctc_decoder_config.vocabulary = ListConfig(vocabulary)
            else:
                ctc_decoder_config.vocabulary = ListConfig(list(vocabulary.keys()))

            decoder_num_classes = ctc_decoder_config['num_classes']
            # Override number of classes if placeholder provided
            logging.info(
                "\nReplacing old number of classes ({}) with new number of classes - {}".format(
                    decoder_num_classes, len(vocabulary)
                )
            )
            ctc_decoder_config['num_classes'] = len(vocabulary)

            del self.ctc_decoder
            self.ctc_decoder = EncDecHybridRNNTCTCBPEModel.from_config_dict(ctc_decoder_config)
            del self.ctc_loss
            self.ctc_loss = CTCLoss(
                num_classes=self.ctc_decoder.num_classes_with_blank - 1,
                zero_infinity=True,
                reduction=self.cfg.aux_ctc.get("ctc_reduction", "mean_batch"),
            )

            if ctc_decoding_cfg is None:
                # Assume same decoding config as before
                ctc_decoding_cfg = self.cfg.aux_ctc.decoding

            # Assert the decoding config with all hyper parameters
            ctc_decoding_cls = OmegaConf.structured(CTCBPEDecodingConfig)
            ctc_decoding_cls = OmegaConf.create(OmegaConf.to_container(ctc_decoding_cls))
            ctc_decoding_cfg = OmegaConf.merge(ctc_decoding_cls, ctc_decoding_cfg)

            self.ctc_decoding = CTCBPEDecoding(decoding_cfg=ctc_decoding_cfg, tokenizer=self.tokenizer)

            self.ctc_wer = WER(
                decoding=self.ctc_decoding,
                use_cer=self.cfg.aux_ctc.get('use_cer', False),
                log_prediction=self.cfg.get("log_prediction", False),
                dist_sync_on_step=True,
            )

            # Update config
            with open_dict(self.cfg.aux_ctc):
                self.cfg.aux_ctc.decoder = ctc_decoder_config

            with open_dict(self.cfg.aux_ctc):
                self.cfg.aux_ctc.decoding = ctc_decoding_cfg

            logging.info(f"Changed tokenizer of the CTC decoder to {self.ctc_decoder.vocabulary} vocabulary.")

    def change_decoding_strategy(
        self, decoding_cfg: DictConfig = None, decoder_type: str = None, verbose: bool = True
    ):
        """
        Changes decoding strategy used during RNNT decoding process.
        Args:
            decoding_cfg: A config for the decoder, which is optional. If the decoding type
                needs to be changed (from say Greedy to Beam decoding etc), the config can be passed here.
            decoder_type: (str) Can be set to 'rnnt' or 'ctc' to switch between appropriate decoder in a
                model having both RNN-T and CTC decoders. Defaults to None, in which case RNN-T decoder is
                used. If set to 'ctc', it raises error if 'ctc_decoder' is not an attribute of the model.
            verbose: bool whether to display change of decoder config or not.
        """
        if decoder_type is None or decoder_type == 'rnnt':
            if decoding_cfg is None:
                # Assume same decoding config as before
                logging.info("No `decoding_cfg` passed when changing decoding strategy, using internal config")
                decoding_cfg = self.cfg.decoding

            # Assert the decoding config with all hyper parameters
            decoding_cls = OmegaConf.structured(RNNTBPEDecodingConfig)
            decoding_cls = OmegaConf.create(OmegaConf.to_container(decoding_cls))
            decoding_cfg = OmegaConf.merge(decoding_cls, decoding_cfg)
            decoding_cfg = self.set_decoding_type_according_to_loss(decoding_cfg)

            self.decoding = RNNTBPEDecoding(
                decoding_cfg=decoding_cfg,
                decoder=self.decoder,
                joint=self.joint,
                tokenizer=self.tokenizer,
            )

            self.wer = WER(
                decoding=self.decoding,
                batch_dim_index=self.wer.batch_dim_index,
                use_cer=self.wer.use_cer,
                log_prediction=self.wer.log_prediction,
                dist_sync_on_step=True,
            )

            # Setup fused Joint step
            if self.joint.fuse_loss_wer or (
                self.decoding.joint_fused_batch_size is not None and self.decoding.joint_fused_batch_size > 0
            ):
                self.joint.set_loss(self.loss)
                self.joint.set_wer(self.wer)

            self.joint.temperature = decoding_cfg.get('temperature', 1.0)

            # Update config
            with open_dict(self.cfg.decoding):
                self.cfg.decoding = decoding_cfg

            self.cur_decoder = "rnnt"
            if verbose:
                logging.info(
                    f"Changed decoding strategy of the RNNT decoder to \n{OmegaConf.to_yaml(self.cfg.decoding)}"
                )

        elif decoder_type == 'ctc':
            if not hasattr(self, 'ctc_decoding'):
                raise ValueError("The model does not have the ctc_decoding module and does not support ctc decoding.")
            if decoding_cfg is None:
                # Assume same decoding config as before
                logging.info("No `decoding_cfg` passed when changing decoding strategy, using internal config")
                decoding_cfg = self.cfg.aux_ctc.decoding

            # Assert the decoding config with all hyper parameters
            decoding_cls = OmegaConf.structured(CTCBPEDecodingConfig)
            decoding_cls = OmegaConf.create(OmegaConf.to_container(decoding_cls))
            decoding_cfg = OmegaConf.merge(decoding_cls, decoding_cfg)

            self.ctc_decoding = CTCBPEDecoding(decoding_cfg=decoding_cfg, tokenizer=self.tokenizer)

            self.ctc_wer = WER(
                decoding=self.ctc_decoding,
                use_cer=self.ctc_wer.use_cer,
                log_prediction=self.ctc_wer.log_prediction,
                dist_sync_on_step=True,
            )

            self.ctc_decoder.temperature = decoding_cfg.get('temperature', 1.0)

            # Update config
            with open_dict(self.cfg.aux_ctc.decoding):
                self.cfg.aux_ctc.decoding = decoding_cfg

            self.cur_decoder = "ctc"
            if verbose:
                logging.info(
                    f"Changed decoding strategy of the CTC decoder to \n{OmegaConf.to_yaml(self.cfg.aux_ctc.decoding)}"
                )
        else:
            raise ValueError(f"decoder_type={decoder_type} is not supported. Supported values: [ctc,rnnt]")

    @classmethod
    def list_available_models(cls) -> List[PretrainedModelInfo]:
        """
        This method returns a list of pre-trained model which can be instantiated directly from NVIDIA's NGC cloud.

        Returns:
            List of available pre-trained models.
        """
        results = []

        model = PretrainedModelInfo(
            pretrained_model_name="stt_en_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_en_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_en_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_en_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_de_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_de_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_de_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_de_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_it_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_it_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_it_fastconformer_hybrid_large_pc/versions/1.20.0/files/stt_it_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_es_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_es_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_es_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_es_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_hr_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_hr_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_hr_fastconformer_hybrid_large_pc/versions/1.21.0/files/FastConformer-Hybrid-Transducer-CTC-BPE-v256-averaged.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_ua_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_ua_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_ua_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_ua_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_pl_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_pl_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_pl_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_pl_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_by_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_by_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_by_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_by_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_ru_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_ru_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_ru_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_ru_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_fr_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_fr_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_fr_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_fr_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_multilingual_fastconformer_hybrid_large_pc_blend_eu",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_multilingual_fastconformer_hybrid_large_pc_blend_eu",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_multilingual_fastconformer_hybrid_large_pc_blend_eu/versions/1.21.0/files/stt_multilingual_fastconformer_hybrid_large_pc_blend_eu.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_multilingual_fastconformer_hybrid_large_pc",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_multilingual_fastconformer_hybrid_large_pc",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_multilingual_fastconformer_hybrid_large_pc/versions/1.21.0/files/stt_multilingual_fastconformer_hybrid_large_pc.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_en_fastconformer_hybrid_large_streaming_80ms",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_en_fastconformer_hybrid_large_streaming_80ms",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_en_fastconformer_hybrid_large_streaming_80ms/versions/1.20.0/files/stt_en_fastconformer_hybrid_large_streaming_80ms.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_en_fastconformer_hybrid_large_streaming_480ms",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_en_fastconformer_hybrid_large_streaming_480ms",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_en_fastconformer_hybrid_large_streaming_480ms/versions/1.20.0/files/stt_en_fastconformer_hybrid_large_streaming_480ms.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_en_fastconformer_hybrid_large_streaming_1040ms",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_en_fastconformer_hybrid_large_streaming_1040ms",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_en_fastconformer_hybrid_large_streaming_1040ms/versions/1.20.0/files/stt_en_fastconformer_hybrid_large_streaming_1040ms.nemo",
        )
        results.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="stt_en_fastconformer_hybrid_large_streaming_multi",
            description="For details about this model, please visit https://ngc.nvidia.com/catalog/models/nvidia:nemo:stt_en_fastconformer_hybrid_large_streaming_multi",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_en_fastconformer_hybrid_large_streaming_multi/versions/1.20.0/files/stt_en_fastconformer_hybrid_large_streaming_multi.nemo",
        )
        results.append(model)

        return results



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch
from omegaconf import DictConfig

from nemo.collections.asr.losses import BCELoss
from nemo.collections.asr.models import EncDecDiarLabelModel


@pytest.fixture()
def msdd_model():

    preprocessor = {
        'cls': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor',
        'params': {
            "features": 80,
            "window_size": 0.025,
            "window_stride": 0.01,
            "sample_rate": 16000,
        },
    }

    speaker_model_encoder = {
        'cls': 'nemo.collections.asr.modules.ConvASREncoder',
        'params': {
            'feat_in': 80,
            'activation': 'relu',
            'conv_mask': True,
            'jasper': [
                {
                    'filters': 512,
                    'repeat': 1,
                    'kernel': [1],
                    'stride': [1],
                    'dilation': [1],
                    'dropout': 0.0,
                    'residual': False,
                    'separable': False,
                }
            ],
        },
    }

    speaker_model_decoder = {
        'cls': 'nemo.collections.asr.modules.SpeakerDecoder',
        'params': {'feat_in': 512, 'num_classes': 2, 'pool_mode': 'xvector', 'emb_sizes': [1024]},
    }

    speaker_model_cfg = DictConfig(
        {
            'preprocessor': DictConfig(preprocessor),
            'encoder': DictConfig(speaker_model_encoder),
            'decoder': DictConfig(speaker_model_decoder),
        }
    )

    msdd_module = {
        'cls': 'nemo.collections.asr.modules.MSDD_module',
        'params': {
            "num_spks": 2,
            "hidden_size": 256,
            "num_lstm_layers": 3,
            "dropout_rate": 0.5,
            "cnn_output_ch": 32,
            "conv_repeat": 2,
            "emb_dim": 192,
            "scale_n": 5,
            "weighting_scheme": 'conv_scale_weight',
            "context_vector_type": 'cos_sim',
        },
    }

    loss = {'cls': 'nemo.collections.asr.losses.bce_loss.BCELoss', 'params': {"weight": None}}

    diarizer = {
        'out_dir': None,
        'oracle_vad': True,
        "speaker_embeddings": {
            "model_path": None,
            "parameters": {
                "window_length_in_sec": [1.5, 1.25, 1.0, 0.75, 0.5],
                "shift_length_in_sec": [0.75, 0.625, 0.5, 0.375, 0.25],
                "multiscale_weights": [1, 1, 1, 1, 1],
                "save_embeddings": True,
            },
        },
    }

    modelConfig = DictConfig(
        {
            'msdd_module': DictConfig(msdd_module),
            'preprocessor': DictConfig(preprocessor),
            'diarizer': DictConfig(diarizer),
            'loss': DictConfig(loss),
            'max_num_of_spks': 2,
            'num_workers': 5,
            'emb_batch_size': 0,
            'soft_label_thres': 0.5,
            'scale_n': 5,
            'speaker_model_cfg': speaker_model_cfg,
        }
    )
    model = EncDecDiarLabelModel(cfg=modelConfig)
    return model


class TestEncDecDiarLabelModel:
    @pytest.mark.unit
    def test_constructor(self, msdd_model):
        diar_model = msdd_model.train()
        assert diar_model.cfg.scale_n == len(
            diar_model.cfg.diarizer.speaker_embeddings.parameters.window_length_in_sec
        )
        assert diar_model.cfg.scale_n == len(diar_model.cfg.diarizer.speaker_embeddings.parameters.shift_length_in_sec)
        assert diar_model.cfg.scale_n == len(diar_model.cfg.diarizer.speaker_embeddings.parameters.multiscale_weights)
        assert diar_model.cfg.msdd_module.num_spks == diar_model.cfg.max_num_of_spks
        # TODO: make proper config and assert correct number of weights
        # Check to/from config_dict:
        confdict = diar_model.to_config_dict()
        instance2 = EncDecDiarLabelModel.from_config_dict(confdict)
        assert isinstance(instance2, EncDecDiarLabelModel)

    @pytest.mark.unit
    def test_forward_infer(self, msdd_model):
        diar_model = msdd_model.eval()

        # batch_size 4, scale_n 5, length 25, emb_dim 192
        input_signal = torch.randn(size=(4, 25, 5, 192))
        input_signal_length = 25 * torch.ones(4, dtype=torch.int)
        emb_vectors = torch.randn(size=(4, 5, 192, 2))
        targets = torch.randint(2, size=(4, 25, 2), dtype=torch.int)

        with torch.no_grad():
            # batch size 1
            preds_list, scale_weights_list = [], []
            for i in range(input_signal.size(0)):
                preds, scale_weights = diar_model.forward_infer(
                    input_signal[i : i + 1], input_signal_length[i : i + 1], emb_vectors[i : i + 1], targets[i : i + 1]
                )
                preds_list.append(preds)
                scale_weights_list.append(scale_weights)
            preds_instance = torch.cat(preds_list, 0)
            scale_weights_instance = torch.cat(scale_weights_list, 0)

            # batch size 4
            preds_batch, scale_weights_batch = diar_model.forward_infer(
                input_signal, input_signal_length, emb_vectors, targets
            )

        assert preds_instance.shape == preds_batch.shape
        assert scale_weights_instance.shape == scale_weights_batch.shape

        diff = torch.mean(torch.abs(preds_instance - preds_batch))
        assert diff <= 1e-6
        diff = torch.max(torch.abs(preds_instance - preds_batch))
        assert diff <= 1e-6
        diff = torch.mean(torch.abs(scale_weights_instance - scale_weights_batch))
        assert diff <= 1e-6
        diff = torch.max(torch.abs(scale_weights_instance - scale_weights_batch))
        assert diff <= 1e-6


class TestBCELoss:
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "probs, labels, target_lens, reduction, expected_output",
        [
            (
                torch.tensor([[[0.5, 0.5], [0.5, 0.5]]], dtype=torch.float32),
                torch.tensor([[[1, 0], [0, 1]]], dtype=torch.float32),
                torch.tensor([2]),
                "mean",
                torch.tensor(0.693147, dtype=torch.float32),
            ),
            (
                torch.tensor([[[0.5, 0.5], [0.0, 1.0]]], dtype=torch.float32),
                torch.tensor([[[1, 0], [0, 1]]], dtype=torch.float32),
                torch.tensor([1]),
                "mean",
                torch.tensor(0.693147, dtype=torch.float32),
            ),
            (
                torch.tensor([[[0, 1], [1, 0]]], dtype=torch.float32),
                torch.tensor([[[1, 0], [0, 1]]], dtype=torch.float32),
                torch.tensor([2]),
                "mean",
                torch.tensor(100, dtype=torch.float32),
            ),
        ],
    )
    def test_loss(self, probs, labels, target_lens, reduction, expected_output):
        loss = BCELoss(reduction=reduction)
        result = loss(probs=probs, labels=labels, target_lens=target_lens)
        assert torch.allclose(result, expected_output), f"Expected {expected_output}, but got {result}"



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import torch
from torch import nn

from nemo.core.classes import Serialization, Typing, typecheck
from nemo.core.neural_types import LabelsType, LogitsType, LossType, MaskType, NeuralType

__all__ = ["BCEWithLogitsLoss"]


class BCEWithLogitsLoss(nn.BCEWithLogitsLoss, Serialization, Typing):
    """
    BCEWithLogitsLoss

    https://pytorch.org/docs/1.9.1/generated/torch.nn.BCEWithLogitsLoss.html
    """

    @property
    def input_types(self):
        """Returns definitions of module input ports.
        """
        return {
            "logits": NeuralType(["B"] + ["ANY"] * (self._logits_dim - 1), LogitsType()),
            "labels": [NeuralType(["B"] + ["ANY"] * (self._logits_dim - 2), LabelsType())],
            "loss_mask": NeuralType(["B"] + ["ANY"] * (self._logits_dim - 2), MaskType(), optional=True),
        }

    @property
    def output_types(self):
        """Returns definitions of module output ports.
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(
        self,
        logits_ndim: int = 2,
        weight: torch.Tensor = None,
        reduction: str = "mean",
        pos_weight: torch.Tensor = None,
    ):
        """
        Args:
            logits_ndim: number of dimensions (or rank) of the logits tensor
            weight: list of rescaling weight given to each class
            reduction: type of the reduction over the batch
            pos_weight: weight given to positive samples
        """
        if pos_weight is not None and not torch.is_tensor(pos_weight):
            pos_weight = torch.FloatTensor(pos_weight)

        super().__init__(weight=weight, pos_weight=pos_weight, reduction=reduction)
        self._logits_dim = logits_ndim

    @typecheck()
    def forward(self, logits: float, labels: List[int], loss_mask: torch.Tensor = None):
        """
        Args:
            logits: output of the classifier
            labels: ground truth labels
        """
        labels = torch.stack(labels)
        labels = labels.t().float()

        return super().forward(logits, labels)



# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# Copyright (c) 2024 Arc Institute. All rights reserved.
# Copyright (c) 2024 Michael Poli. All rights reserved.
# Copyright (c) 2024 Stanford University. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lightning.pytorch as pl
import nemo_run as run

from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec
from nemo.collections.llm.api import finetune, pretrain
from nemo.collections.llm.recipes.hyena_base import model_recipe, pretrain_recipe_creater, tokenizer_recipe


NAME = "hyena_1b"


@run.cli.factory(name=NAME)
def tokenizer() -> run.Config[TokenizerSpec]:
    """
    Defines a factory function for creating a tokenizer configuration.

    This function is registered as a CLI factory with the specified name and
    returns a tokenizer configuration based on the `tokenizer_recipe`.

    Returns:
        run.Config[TokenizerSpec]: A configuration object for the tokenizer.
    """
    return tokenizer_recipe()


@run.cli.factory(name=NAME)
def model(tp_comm_overlap: bool = False, seq_length: int = 8192) -> run.Config[pl.LightningModule]:
    """
    Factory function to create a Striped-Hyena 1B model configuration.

    Returns:
        run.Config[pl.LightningModule]: Configuration for the Striped-Hyena 1B model.
    """
    return model_recipe('1b', tp_comm_overlap, seq_length)


@run.cli.factory(target=pretrain, name=NAME)
def pretrain_recipe(
    dir=None,
    micro_batch_size=1,
    global_batch_size=8,
    num_nodes=1,
    num_gpus_per_node=8,
    tensor_parallel_size=1,
    context_parallel_size=1,
    model_size='1b',
    fn=pretrain,
    **kwargs,
) -> run.Partial:
    """
    Create a pre-training recipe for Striped-Hyena 1B model.

    This function sets up a complete configuration for pre-training, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the pre-training run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        fn (Callable): The pre-training function to use.

    Returns:
        run.Partial: Partial configuration for pre-training.

    """
    return pretrain_recipe_creater(
        dir=dir,
        micro_batch_size=micro_batch_size,
        global_batch_size=global_batch_size,
        num_nodes=num_nodes,
        num_gpus_per_node=num_gpus_per_node,
        tensor_parallel_size=tensor_parallel_size,
        context_parallel_size=context_parallel_size,
        model_size=model_size,
        **kwargs,
    )


@run.cli.factory(target=finetune, name=NAME)
def finetune_recipe(
    dir=None,
    resume_path="hyena-1b-pretrain",
    micro_batch_size=1,
    global_batch_size=8,
    num_nodes=1,
    num_gpus_per_node=8,
    tensor_parallel_size=1,
    context_parallel_size=1,
    model_size='1b',
    fn=finetune,
    **kwargs,
) -> run.Partial:
    """ """

    return pretrain_recipe_creater(
        dir=dir,
        resume_path=resume_path,
        micro_batch_size=micro_batch_size,
        global_batch_size=global_batch_size,
        num_nodes=num_nodes,
        num_gpus_per_node=num_gpus_per_node,
        tensor_parallel_size=tensor_parallel_size,
        context_parallel_size=context_parallel_size,
        model_size=model_size,
        fn=fn,
        **kwargs,
    )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os.path
import re
from functools import cached_property
from typing import Any

from nemo.collections.asr.models import ASRModel
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis
from nemo.collections.asr.parts.utils.timestamp_utils import get_segment_offsets, get_words_offsets


class BaseTimestampsTest:
    """
    Base class for testing timestamps in decoders (CTC and RNNT).
    This class defines common test methods that can be inherited by both
    test_ctc_decoding.py and test_rnnt_decoding.py.
    """

    @cached_property
    def bpe_tokenizer(self):
        model_path = "/home/TestData/asr/stt_en_conformer_transducer_small.nemo"
        if os.path.exists(model_path):
            model = ASRModel.restore_from(model_path, map_location="cpu")
        else:
            model = ASRModel.from_pretrained("stt_en_conformer_transducer_small", map_location="cpu")
        return model.tokenizer

    @property
    def char_offsets_chars(self):
        char_offsets = [
            {"char": "e", "start_offset": 0, "end_offset": 1},
            {"char": " ", "start_offset": 2, "end_offset": 2},
            {"char": "e", "start_offset": 3, "end_offset": 4},
            {"char": " ", "start_offset": 5, "end_offset": 5},
            {"char": ".", "start_offset": 6, "end_offset": 7},
            {"char": " ", "start_offset": 8, "end_offset": 9},
            {"char": "e", "start_offset": 10, "end_offset": 11},
            {"char": " ", "start_offset": 12, "end_offset": 13},
            {"char": "?", "start_offset": 14, "end_offset": 15},
            {"char": " ", "start_offset": 16, "end_offset": 17},
        ]
        return char_offsets

    @property
    def word_offsets_chars_expected_output(self):
        return [
            {'word': 'e', 'start_offset': 0, 'end_offset': 1},
            {'word': 'e.', 'start_offset': 3, 'end_offset': 7},
            {'word': 'e?', 'start_offset': 10, 'end_offset': 15},
        ]

    @property
    def word_offsets_chars_expected_output_other_delimiter(self):
        return [
            {'word': 'e e ', 'start_offset': 0, 'end_offset': 5},
            {'word': ' e? ', 'start_offset': 8, 'end_offset': 17},
        ]

    @property
    def segment_offsets_expected_output(self):
        return [
            {'segment': 'e e.', 'start_offset': 0, 'end_offset': 7},
            {'segment': 'e?', 'start_offset': 10, 'end_offset': 15},
        ]

    @property
    def segment_offsets_expected_output_gap(self):
        return [
            {'segment': 'e e. e?', 'start_offset': 0, 'end_offset': 15},
        ]

    @property
    def char_offsets_wpe(self):
        char_offsets = [
            {"char": "nineteen", "start_offset": 0, "end_offset": 1},
            {"char": "##th", "start_offset": 2, "end_offset": 2},
            {"char": "re", "start_offset": 3, "end_offset": 4},
            {"char": "seven", "start_offset": 5, "end_offset": 6},
            {"char": "##ty", "start_offset": 6, "end_offset": 7},
            {"char": "eighty", "start_offset": 8, "end_offset": 9},
        ]

        return char_offsets

    @property
    def word_offsets_wpe_expected_output(self):
        return [
            {'word': 'nineteenth', 'start_offset': 0, 'end_offset': 2},
            {'word': 're', 'start_offset': 3, 'end_offset': 4},
            {'word': 'seventy', 'start_offset': 5, 'end_offset': 7},
            {'word': 'eighty', 'start_offset': 8, 'end_offset': 9},
        ]

    @property
    def word_offsets_wpe_expected_output_other_delimiter(self):
        return [
            {'word': 'nineteenth', 'start_offset': 0, 'end_offset': 2},
            {'word': 'seventy eighty', 'start_offset': 5, 'end_offset': 9},
        ]

    @property
    def char_offsets_bpe(self):
        char_offsets = [
            {"char": "discuss", "start_offset": 0, "end_offset": 2},
            {"char": "absolute", "start_offset": 2, "end_offset": 4},
            {"char": "'", "start_offset": 5, "end_offset": 5},
            {"char": "really", "start_offset": 5, "end_offset": 6},
            {"char": "friend", "start_offset": 6, "end_offset": 7},
            {"char": "ship", "start_offset": 8, "end_offset": 9},
        ]

        return char_offsets

    @property
    def encoded_char_offsets_bpe(self):
        char_offsets = [
            {"char": "▁discuss", "start_offset": 0, "end_offset": 2},
            {"char": "▁absolute", "start_offset": 2, "end_offset": 4},
            {"char": "'", "start_offset": 5, "end_offset": 5},
            {"char": "▁really", "start_offset": 5, "end_offset": 6},
            {"char": "▁friend", "start_offset": 6, "end_offset": 7},
            {"char": "ship", "start_offset": 8, "end_offset": 9},
        ]

        return char_offsets

    @property
    def word_offsets_bpe_expected_output(self):
        return [
            {'word': "discuss", 'start_offset': 0, 'end_offset': 2},
            {'word': "absolute'", 'start_offset': 2, 'end_offset': 5},
            {'word': "really", 'start_offset': 5, 'end_offset': 6},
            {'word': "friendship", 'start_offset': 6, 'end_offset': 9},
        ]

    @property
    def word_offsets_bpe_expected_output_other_delimiter(self):
        return [
            {'word': "discuss absolute'", 'start_offset': 0, 'end_offset': 5},
            {'word': "friendship", 'start_offset': 6, 'end_offset': 9},
        ]

    @staticmethod
    def check_char_timestamps(hyp: Hypothesis, decoding: Any):
        """Test character-level timestamps for both CTC and RNNT"""
        assert hyp.timestamp is not None
        assert isinstance(hyp.timestamp, dict)
        assert 'timestep' in hyp.timestamp
        assert 'char' in hyp.timestamp
        assert 'word' in hyp.timestamp
        assert 'segment' in hyp.timestamp

        hypothesis_text = re.sub(r'\s+', ' ', hyp.text.strip())

        words = hyp.text.split(decoding.word_seperator)
        words = list(filter(lambda x: x != '', words))
        assert len(hyp.timestamp['word']) == len(words)

        words_from_timestamps = [ts['word'] for ts in hyp.timestamp['word']]
        assert hypothesis_text == decoding.word_seperator.join(words_from_timestamps)

        segments = []
        segment = []

        for word in words:
            segment.append(word)
            if word[-1] in decoding.segment_seperators:
                segments.append(' '.join(segment))
                segment = []

        if segment:
            segments.append(' '.join(segment))

        assert len(hyp.timestamp['segment']) == len(segments)

        segments_from_timestamps = [ts['segment'] for ts in hyp.timestamp['segment']]
        assert hypothesis_text == decoding.word_seperator.join(segments_from_timestamps)

    @staticmethod
    def check_subword_timestamps(hyp: Hypothesis, decoding: Any):
        """Test subword-level timestamps for both CTC and RNNT"""
        assert hyp.timestamp is not None
        assert isinstance(hyp.timestamp, dict)
        assert 'timestep' in hyp.timestamp
        assert 'char' in hyp.timestamp
        assert 'word' in hyp.timestamp
        assert 'segment' in hyp.timestamp

        chars = list(hyp.text)
        chars = list(filter(lambda x: x not in ['', ' ', '#'], chars))
        all_chars = [list(decoding.tokenizer.tokens_to_text(data['char'])) for data in hyp.timestamp['char']]
        all_chars = [char for subword in all_chars for char in subword]
        all_chars = list(filter(lambda x: x not in ['', ' ', '#'], all_chars))
        assert len(chars) == len(all_chars)

        hypothesis_text = re.sub(r'\s+', ' ', hyp.text.strip())

        words_from_timestamps = [ts['word'] for ts in hyp.timestamp['word']]
        assert hypothesis_text == decoding.word_seperator.join(words_from_timestamps)

        segments_count = sum([hyp.text.count(seperator) for seperator in decoding.segment_seperators])
        if hyp.text[-1] not in decoding.segment_seperators:
            segments_count += 1

        if hyp.text in decoding.segment_seperators:
            segments_count = 0

        assert len(hyp.timestamp['segment']) == segments_count

        segments_from_timestamps = [ts['segment'] for ts in hyp.timestamp['segment']]
        assert hypothesis_text == decoding.word_seperator.join(segments_from_timestamps)

    def test_word_offsets_chars(self):
        word_offsets = get_words_offsets(
            char_offsets=self.char_offsets_chars,
            encoded_char_offsets=None,
            word_delimiter_char=" ",
            tokenizer_type="char",
            supported_punctuation={'.', '!', '?'},
            decode_tokens_to_str=self.decoding_char.decode_tokens_to_str,
        )

        assert word_offsets == self.word_offsets_chars_expected_output

    def test_word_offsets_char_other_delimiter(self):
        word_offsets = get_words_offsets(
            char_offsets=self.char_offsets_chars,
            encoded_char_offsets=None,
            tokenizer_type="char",
            word_delimiter_char=".",
            supported_punctuation={'.', '!', '?'},
            decode_tokens_to_str=self.decoding_char.decode_tokens_to_str,
        )

        assert word_offsets == self.word_offsets_chars_expected_output_other_delimiter

    def test_word_offsets_subword_wpe(self):

        word_offsets = get_words_offsets(
            char_offsets=self.char_offsets_wpe,
            encoded_char_offsets=None,
            word_delimiter_char=" ",
            tokenizer_type="wpe",
            supported_punctuation={'.', '!', '?'},
            decode_tokens_to_str=self.decoding_subword_wpe.decode_tokens_to_str,
        )

        assert word_offsets == self.word_offsets_wpe_expected_output

    def test_word_offsets_subword_wpe_other_delimiter(self):

        word_offsets = get_words_offsets(
            char_offsets=self.char_offsets_wpe,
            encoded_char_offsets=None,
            word_delimiter_char="re",
            tokenizer_type="wpe",
            supported_punctuation={'.', '!', '?'},
            decode_tokens_to_str=self.decoding_subword_wpe.decode_tokens_to_str,
        )

        assert word_offsets == self.word_offsets_wpe_expected_output_other_delimiter

    def test_word_offsets_subword_bpe(self):

        word_offsets = get_words_offsets(
            char_offsets=self.char_offsets_bpe,
            encoded_char_offsets=self.encoded_char_offsets_bpe,
            word_delimiter_char=" ",
            tokenizer_type="bpe",
            supported_punctuation={'.', '!', '?'},
            decode_tokens_to_str=self.decoding_subword_bpe.decode_tokens_to_str,
        )

        assert word_offsets == self.word_offsets_bpe_expected_output

    def test_word_offsets_subword_bpe_other_delimiter(self):
        word_offsets = get_words_offsets(
            char_offsets=self.char_offsets_bpe,
            encoded_char_offsets=self.encoded_char_offsets_bpe,
            word_delimiter_char="really",
            tokenizer_type="bpe",
            supported_punctuation={'.', '!', '?'},
            decode_tokens_to_str=self.decoding_subword_bpe.decode_tokens_to_str,
        )

        assert word_offsets == self.word_offsets_bpe_expected_output_other_delimiter

    def test_segment_offsets_delimiter(self):
        segment_offsets = get_segment_offsets(
            word_offsets=self.word_offsets_chars_expected_output,
            segment_delimiter_tokens=['.', '!', '?'],
            supported_punctuation={'.', '!', '?'},
        )

        assert segment_offsets == self.segment_offsets_expected_output

    def test_segment_offsets_gap(self):
        segment_offsets = get_segment_offsets(
            word_offsets=self.word_offsets_chars_expected_output,
            segment_delimiter_tokens=[],
            supported_punctuation={},
            segment_gap_threshold=10,
        )

        assert segment_offsets == self.segment_offsets_expected_output_gap



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from abc import ABC, abstractmethod
from typing import Optional

import torch

from nemo.collections.common.parts.utils import mask_sequence_tensor
from nemo.core.classes import NeuralModule, typecheck
from nemo.core.neural_types import LengthsType, NeuralType, SpectrogramType
from nemo.utils import logging


class SBNoiseSchedule(NeuralModule, ABC):
    """Noise schedule for the Schrödinger Bridge

    Args:
        time_min: minimum time for the process
        time_max: maximum time for the process
        num_steps: number of steps for the process
        eps: small regularization

    References:
        Schrödinger Bridge for Generative Speech Enhancement, https://arxiv.org/abs/2407.16074
    """

    def __init__(
        self,
        time_min: float = 0.0,
        time_max: float = 1.0,
        num_steps: int = 100,
        eps: float = 1e-8,
    ):
        super().__init__()

        # min and max time
        if time_min < 0:
            raise ValueError(f'time_min should be non-negative, current value {time_min}')

        if time_max <= time_min:
            raise ValueError(f'time_max should be larger than time_min, current max {time_max} and min {time_min}')

        self.time_min = time_min
        self.time_max = time_max

        if num_steps <= 0:
            raise ValueError(f'Expected num_steps > 0, got {num_steps}')

        self.num_steps = num_steps

        if eps <= 0:
            raise ValueError(f'Expected eps > 0, got {eps}')

        self.eps = eps

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\ttime_min:  %s', self.time_min)
        logging.debug('\ttime_max:  %s', self.time_max)
        logging.debug('\tnum_steps: %s', self.num_steps)
        logging.debug('\teps:       %s', self.eps)

    @property
    def dt(self) -> float:
        """Time step for the process."""
        return self.time_max / self.num_steps

    @property
    def time_delta(self) -> float:
        """Time range for the process."""
        return self.time_max - self.time_min

    def generate_time(self, size: int, device: torch.device) -> torch.Tensor:
        """Generate random time steps in the valid range."""
        time = torch.rand(size, device=device) * self.time_delta + self.time_min
        return time

    @property
    def alpha_t_max(self):
        """Return alpha_t at t_max."""
        t_max = torch.tensor([self.time_max], device=alpha.device)
        return self.alpha(t_max)

    @property
    def sigma_t_max(self):
        """Return sigma_t at t_max."""
        t_max = torch.tensor([self.time_max], device=alpha.device)
        return self.sigma(t_max)

    @abstractmethod
    def f(self, time: torch.Tensor) -> torch.Tensor:
        """Drift scaling f(t).

        Args:
            time: tensor with time steps

        Returns:
            Tensor the same size as time, representing drift scaling.
        """
        pass

    @abstractmethod
    def g(self, time: torch.Tensor) -> torch.Tensor:
        """Diffusion scaling g(t).

        Args:
            time: tensor with time steps

        Returns:
            Tensor the same size as time, representing diffusion scaling.
        """
        pass

    @abstractmethod
    def alpha(self, time: torch.Tensor) -> torch.Tensor:
        """Return alpha for SB noise schedule.

            alpha_t = exp( int_0^s f(s) ds  )

        Args:
            time: tensor with time steps

        Returns:
            Tensor the same size as time, representing alpha for each time.
        """
        pass

    def alpha_bar_from_alpha(self, alpha: torch.Tensor) -> (torch.Tensor, torch.Tensor):
        """Return alpha_bar for SB.

            alpha_bar = alpha_t / alpha_t_max

        Args:
            alpha: tensor with alpha values

        Returns:
            Tensors the same size as alpha, representing alpha_bar and alpha_t_max.
        """
        alpha_t_max = self.alpha(torch.tensor([self.time_max], device=alpha.device))
        alpha_bar = alpha / (alpha_t_max + self.eps)
        return alpha_bar, alpha_t_max

    def get_alphas(self, time: torch.Tensor) -> (torch.Tensor, torch.Tensor, torch.Tensor):
        """Return alpha, alpha_bar and alpha_t_max for SB.

        Args:
            time: tensor with time steps

        Returns:
            Tuple of tensors with alpha, alpha_bar and alpha_t_max.
        """
        alpha = self.alpha(time)
        alpha_bar, alpha_t_max = self.alpha_bar_from_alpha(alpha)
        return alpha, alpha_bar, alpha_t_max

    @abstractmethod
    def sigma(self, time: torch.Tensor) -> torch.Tensor:
        """Return sigma_t for SB.

            sigma_t^2 = int_0^s g^2(s) / alpha_s^2 ds

        Args:
            time: tensor with time steps

        Returns:
            Tensor the same size as time, representing sigma for each time.
        """
        pass

    def sigma_bar_from_sigma(self, sigma: torch.Tensor) -> (torch.Tensor, torch.Tensor):
        """Return sigma_bar_t for SB.

            sigma_bar_t^2 = sigma_t_max^2 - sigma_t^2

        Args:
            sigma: tensor with sigma values

        Returns:
            Tensors the same size as sigma, representing sigma_bar and sigma_t_max.
        """
        sigma_t_max = self.sigma(torch.tensor([self.time_max], device=sigma.device))
        sigma_bar_sq = sigma_t_max**2 - sigma**2
        return torch.sqrt(sigma_bar_sq + self.eps), sigma_t_max

    def get_sigmas(self, time: torch.Tensor) -> (torch.Tensor, torch.Tensor, torch.Tensor):
        """Return sigma, sigma_bar and sigma_t_max for SB.

        Args:
            time: tensor with time steps

        Returns:
            Tuple of tensors with sigma, sigma_bar and sigma_t_max.
        """
        sigma = self.sigma(time)
        sigma_bar, sigma_t_max = self.sigma_bar_from_sigma(sigma)
        return sigma, sigma_bar, sigma_t_max

    @abstractmethod
    def copy(self):
        """Return a copy of the noise schedule."""
        pass

    def __repr__(self):
        desc = f'{self.__class__.__name__}(time_min={self.time_min}, time_max={self.time_max}, num_steps={self.num_steps})'
        desc += f'\n\tdt:         {self.dt}'
        desc += f'\n\ttime_delta: {self.time_delta}'
        return desc


class SBNoiseScheduleVE(SBNoiseSchedule):
    """Variance exploding noise schedule for the Schrödinger Bridge.

    Args:
        k: defines the base for the exponential diffusion coefficient
        c: scaling for the diffusion coefficient
        time_min: minimum time for the process
        time_max: maximum time for the process
        num_steps: number of steps for the process
        eps: small regularization

    References:
        Schrödinger Bridge for Generative Speech Enhancement, https://arxiv.org/abs/2407.16074
    """

    def __init__(
        self,
        k: float,
        c: float,
        time_min: float = 0.0,
        time_max: float = 1.0,
        num_steps: int = 100,
        eps: float = 1e-8,
    ):
        super().__init__(time_min=time_min, time_max=time_max, num_steps=num_steps, eps=eps)

        # Shape parameters
        if k <= 1:
            raise ValueError(f'Expected k > 1, got {k}')

        if c <= 0:
            raise ValueError(f'Expected c > 0, got {c}')

        self.c = c
        self.k = k

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tk:         %s', self.k)
        logging.debug('\tc:         %s', self.c)
        logging.debug('\ttime_min:  %s', self.time_min)
        logging.debug('\ttime_max:  %s', self.time_max)
        logging.debug('\tnum_steps: %s', self.num_steps)
        logging.debug('\teps:       %s', self.eps)

    def f(self, time: torch.Tensor) -> torch.Tensor:
        return torch.zeros_like(time)

    def g(self, time: torch.Tensor) -> torch.Tensor:
        return torch.sqrt(self.c) * self.k**self.time

    def alpha(self, time: torch.Tensor) -> torch.Tensor:
        return torch.ones_like(time)

    def sigma(self, time: torch.Tensor) -> torch.Tensor:
        sigma_sq = self.c * (self.k ** (2 * time) - 1) / (2 * math.log(self.k) + self.eps)
        return torch.sqrt(sigma_sq)

    def copy(self):
        return SBNoiseScheduleVE(
            k=self.k,
            c=self.c,
            time_min=self.time_min,
            time_max=self.time_max,
            num_steps=self.num_steps,
            eps=self.eps,
        )

    def __repr__(self):
        desc = super().__repr__()
        desc += f'\n\tk: {self.k}'
        desc += f'\n\tc: {self.c}'
        return desc


class SBNoiseScheduleVP(SBNoiseSchedule):
    """Variance preserving noise schedule for the Schrödinger Bridge.

    Args:
        beta_0: defines the lower bound for diffusion coefficient
        beta_1: defines upper bound for diffusion coefficient
        c: scaling for the diffusion coefficient
        time_min: minimum time for the process
        time_max: maximum time for the process
        num_steps: number of steps for the process
        eps: small regularization
    """

    def __init__(
        self,
        beta_0: float,
        beta_1: float,
        c: float = 1.0,
        time_min: float = 0.0,
        time_max: float = 1.0,
        num_steps: int = 100,
        eps: float = 1e-8,
    ):
        super().__init__(time_min=time_min, time_max=time_max, num_steps=num_steps, eps=eps)

        # Shape parameters
        if beta_0 < 0:
            raise ValueError(f'Expected beta_0 >= 0, got {beta_0}')

        if beta_1 < 0:
            raise ValueError(f'Expected beta_1 >= 0, got {beta_1}')

        if beta_0 >= beta_1:
            raise ValueError(f'Expected beta_0 < beta_1, got beta_0={beta_0} and beta_1={beta_1}')

        if c <= 0:
            raise ValueError(f'Expected c > 0, got {c}')

        self.beta_0 = beta_0
        self.beta_1 = beta_1
        self.c = c

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tbeta_0:    %s', self.beta_0)
        logging.debug('\tbeta_1:    %s', self.beta_1)
        logging.debug('\tc:         %s', self.c)
        logging.debug('\ttime_min:  %s', self.time_min)
        logging.debug('\ttime_max:  %s', self.time_max)
        logging.debug('\tnum_steps: %s', self.num_steps)
        logging.debug('\teps:       %s', self.eps)

    def f(self, time: torch.Tensor) -> torch.Tensor:
        return -0.5 * (self.beta_0 + time * (self.beta_1 - self.beta_0))

    def g(self, time: torch.Tensor) -> torch.Tensor:
        g_sq = self.c * (self.beta_0 + time * (self.beta_1 - self.beta_0))
        return torch.sqrt(g_sq)

    def alpha(self, time: torch.Tensor) -> torch.Tensor:
        tmp = self.beta_0 * time + (self.beta_1 - self.beta_0) / 2 * time**2
        return torch.exp(-0.5 * tmp)

    def sigma(self, time: torch.Tensor) -> torch.Tensor:
        sigma_sq = self.beta_0 * time + (self.beta_1 - self.beta_0) / 2 * time**2
        sigma_sq = torch.exp(sigma_sq) - 1
        sigma_sq = self.c * sigma_sq
        return torch.sqrt(sigma_sq)

    def copy(self):
        return SBNoiseScheduleVP(
            beta_0=self.beta_0,
            beta_1=self.beta_1,
            c=self.c,
            time_min=self.time_min,
            time_max=self.time_max,
            num_steps=self.num_steps,
            eps=self.eps,
        )

    def __repr__(self):
        desc = super().__repr__()
        desc += f'\n\tbeta_0: {self.beta_0}'
        desc += f'\n\tbeta_1: {self.beta_1}'
        desc += f'\n\tc:      {self.c}'
        return desc


class SBSampler(NeuralModule):
    """Schrödinger Bridge sampler.

    Args:
        noise_schedule: noise schedule for the bridge
        estimator: neural estimator
        estimator_output: defines the output of the estimator, e.g., data_prediction
        estimator_time: time for conditioning the estimator, e.g., 'current'
                        or 'previous'. Default is 'previous'.
        process: defines the process, e.g., sde or ode
        time_max: maximum time for the process
        time_min: minimum time for the process
        num_steps: number of steps for the process
        eps: small regularization to prevent division by zero

    References:
        Schrödinger Bridge for Generative Speech Enhancement, https://arxiv.org/abs/2407.16074
        Schrodinger Bridges Beat Diffusion Models on Text-to-Speech Synthesis, https://arxiv.org/abs/2312.03491
    """

    def __init__(
        self,
        noise_schedule: SBNoiseSchedule,
        estimator: NeuralModule,  # neural estimator
        estimator_output: str,
        estimator_time: str = 'previous',  # time for the estimator
        process: str = 'sde',
        time_max: Optional[float] = None,
        time_min: Optional[float] = None,
        num_steps: int = 50,
        eps: float = 1e-8,
    ):
        super().__init__()
        # Create a copy of the noise schedule
        self.noise_schedule = noise_schedule.copy()

        # Update sampling parameters
        if time_max is not None:
            self.noise_schedule.time_max = time_max
            logging.info('noise_schedule.time_max set to: %s', self.noise_schedule.time_max)

        if time_min is not None:
            self.noise_schedule.time_min = time_min
            logging.info('noise_schedule.time_min set to: %s', self.noise_schedule.time_min)

        self.noise_schedule.num_steps = num_steps
        logging.info('noise_schedule.num_steps set to: %s', self.noise_schedule.num_steps)

        # Estimator
        self.estimator = estimator
        self.estimator_output = estimator_output
        self.estimator_time = estimator_time

        # Sampling process
        self.process = process

        # Small regularization
        if eps <= 0:
            raise ValueError(f'Expected eps > 0, got {eps}')
        self.eps = eps

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\testimator_output: %s', self.estimator_output)
        logging.debug('\testimator_time:   %s', self.estimator_time)
        logging.debug('\tprocess:          %s', self.process)
        logging.debug('\ttime_min:         %s', self.time_min)
        logging.debug('\ttime_max:         %s', self.time_max)
        logging.debug('\tnum_steps:        %s', self.num_steps)
        logging.debug('\teps:              %s', self.eps)

    @property
    def time_max(self):
        return self.noise_schedule.time_max

    @time_max.setter
    def time_max(self, value: float):
        self.noise_schedule.time_max = value
        logging.debug('noise_schedule.time_max set to: %s', self.noise_schedule.time_max)

    @property
    def time_min(self):
        return self.noise_schedule.time_min

    @time_min.setter
    def time_min(self, value: float):
        self.noise_schedule.time_min = value
        logging.debug('noise_schedule.time_min set to: %s', self.noise_schedule.time_min)

    @property
    def num_steps(self):
        return self.noise_schedule.num_steps

    @num_steps.setter
    def num_steps(self, value: int):
        self.noise_schedule.num_steps = value
        logging.debug('noise_schedule.num_steps set to: %s', self.noise_schedule.num_steps)

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, value: str):
        if value not in ['sde', 'ode']:
            raise ValueError(f'Unexpected process: {value}')
        self._process = value
        logging.info('process set to: %s', self._process)

    @property
    def estimator_time(self):
        return self._estimator_time

    @estimator_time.setter
    def estimator_time(self, value: str):
        if value not in ['current', 'previous']:
            raise ValueError(f'Unexpected estimator time: {value}')
        self._estimator_time = value
        logging.info('estimator time set to: %s', self._estimator_time)

    @typecheck(
        input_types={
            "prior_mean": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "estimator_condition": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType(), optional=True),
            "state_length": NeuralType(tuple('B'), LengthsType(), optional=True),
        },
        output_types={
            "sample": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "state_length": NeuralType(tuple('B'), LengthsType(), optional=True),
        },
    )
    @torch.inference_mode()
    def forward(
        self, prior_mean: torch.Tensor, estimator_condition: torch.Tensor, state_length: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Takes prior mean and generates a sample."""
        # SB starts from the prior mean
        state = prior_mean

        if state_length is not None:
            state = mask_sequence_tensor(state, state_length)

        # Time steps for sampling
        time_steps = torch.linspace(self.time_max, self.time_min, self.num_steps + 1, device=state.device)

        # Initial values
        time_prev = time_steps[0] * torch.ones(state.shape[0], device=state.device)
        alpha_prev, _, alpha_t_max = self.noise_schedule.get_alphas(time_prev)
        sigma_prev, sigma_bar_prev, sigma_t_max = self.noise_schedule.get_sigmas(time_prev)

        # Sampling
        # Sample at the initial time step (`self.time_max`) is exactly the prior_mean.
        # We do not need to estimate it, but we need to pass it to the next time step.
        # We iterate through the following time steps to generate the sample at the final time (`self.time_min`).
        for t in time_steps[1:]:

            # Prepare time steps for the whole batch
            time = t * torch.ones(state.shape[0], device=state.device)

            # Prepare input for estimator, concatenate conditioning along the channel dimension
            estimator_input = state if estimator_condition is None else torch.cat([state, estimator_condition], dim=1)
            estimator_time = time if self.estimator_time == 'current' else time_prev

            # Estimator
            if self.estimator_output == 'data_prediction':
                current_estimate, _ = self.estimator(
                    input=estimator_input, input_length=state_length, condition=estimator_time
                )
            else:
                raise NotImplementedError(f'Unexpected estimator output: {self.estimator_output}')

            # Get noise schedule for current time
            alpha_t, alpha_bar_t, _ = self.noise_schedule.get_alphas(time)
            sigma_t, sigma_bar_t, _ = self.noise_schedule.get_sigmas(time)

            if self.process == 'sde':
                # Calculate scaling for the first-order discretization from the paper
                weight_prev = alpha_t * sigma_t**2 / (alpha_prev * sigma_prev**2 + self.eps)
                tmp = 1 - sigma_t**2 / (sigma_prev**2 + self.eps)
                weight_estimate = alpha_t * tmp
                weight_z = alpha_t * sigma_t * torch.sqrt(tmp)

                # View as [B, C, D, T]
                weight_prev = weight_prev.view(-1, 1, 1, 1)
                weight_estimate = weight_estimate.view(-1, 1, 1, 1)
                weight_z = weight_z.view(-1, 1, 1, 1)

                # Random sample
                z_norm = torch.randn_like(state)

                # Update state: weighted sum of previous state, current estimate and noise
                state = weight_prev * state + weight_estimate * current_estimate + weight_z * z_norm
            elif self.process == 'ode':
                # Calculate scaling for the first-order discretization from the paper
                weight_prev = alpha_t * sigma_t * sigma_bar_t / (alpha_prev * sigma_prev * sigma_bar_prev + self.eps)
                weight_estimate = (
                    alpha_t
                    / (sigma_t_max**2 + self.eps)
                    * (sigma_bar_t**2 - sigma_bar_prev * sigma_t * sigma_bar_t / (sigma_prev + self.eps))
                )
                weight_prior_mean = (
                    alpha_t
                    / (alpha_t_max * sigma_t_max**2 + self.eps)
                    * (sigma_t**2 - sigma_prev * sigma_t * sigma_bar_t / (sigma_bar_prev + self.eps))
                )

                # View as [B, C, D, T]
                weight_prev = weight_prev.view(-1, 1, 1, 1)
                weight_estimate = weight_estimate.view(-1, 1, 1, 1)
                weight_prior_mean = weight_prior_mean.view(-1, 1, 1, 1)

                # Update state: weighted sum of previous state, current estimate and prior
                state = weight_prev * state + weight_estimate * current_estimate + weight_prior_mean * prior_mean
            else:
                raise RuntimeError(f'Unexpected process: {self.process}')

            # Save previous values
            time_prev = time
            alpha_prev = alpha_t
            sigma_prev = sigma_t
            sigma_bar_prev = sigma_bar_t

        # Final output
        if state_length is not None:
            state = mask_sequence_tensor(state, state_length)

        return state, state_length
